package noown.domain.xsc;

import java.io.Reader;
import java.util.Stack;
import java.util.Vector;

import noown.domain.xsc.comp.*;
import noown.domain.xsc.val.XSCAnnotationValue;
import noown.domain.xsc.val.XSCKeywordValue;
import noown.domain.xsc.val.XSCNameValue;
import noown.domain.xsc.val.XSCSimpleValue;
import noown.domain.xsc.val.XSCValue;

import org.w3c.dom.Attr;
import org.w3c.dom.DOMImplementation;
import org.w3c.dom.Document;
import org.w3c.dom.Element;

/**
 * This class builds a DOM representation in XML Schema syntax
 * from a file containing an XML Schema in compact syntax.
 * 
 * @author Kilian Stillhard
 */
public class XSCDOMBuilder implements XSCHandler
{
        // can only have one, non-reentrant, parser instance
        private static final Object PARSER_LOCK = new Object();
        private static XSCParser parser = null;

	private boolean fDebug;
	private int fIndent;
	private Stack fComponents;
	private XSCComponent fCurrent;
	private Vector fQualifiers;
	private boolean fInComponent;
	private Document fDoc;
	private Vector fAnnotations;
	private boolean fStripAnnots;
	private XSCSchema fSchema;
	
	private static XSCDOMBuilder fBuilder;
	
	public static XSCDOMBuilder getBuilder()
	{
		return fBuilder;
	}
	
	/**
	 * Constructs an XSCDOMBuilder.
	 * 
	 * @param debug When set to true, the events emitted by XSCParser are
	 * logged to System.out.
	 * @param stripAnnots When set to true, all annotations are ignored.
	 */
	public XSCDOMBuilder(boolean debug, boolean stripAnnots)
	{
		fDebug = debug;
		fStripAnnots = stripAnnots;
		fIndent = 0;
		fInComponent = false;
		fCurrent = null;
		fQualifiers = new Vector();
		fComponents = new Stack();
		fAnnotations = new Vector();
		
		fBuilder = this;
	}

	/**
	 * Parse a file in XML Schema compact syntax using the given reader.
	 * 
	 * @param reader The source of the compact syntax document, will be passed
	 * to XSCParser.
	 * @param impl The DOM Implementation to use
	 */
	public Document parse(Reader reader, DOMImplementation impl) throws XSCException
	{
		// initialization
		XSCNamespace.initNamespaces();
		
		// note: the prefix of the schema element will be adapted later
		fDoc = impl.createDocument(
								XSCNamespace.SCHEMA_NS,
								XSCNamespace.getSchemaName("schema"),
								null);
		fSchema = new XSCSchema();
		fCurrent = fSchema;
		
		// the parsing
                synchronized (PARSER_LOCK) {
                    if (parser == null) {
                        parser = new XSCParser(reader);
                    }
                    XSCParser.ReInit(reader);
                    XSCParser.setHandler(this);
                    XSCParser.start();
                    fSchema.freeze(); //PENDING: could this be outside the lock?
                }
		
		return fDoc;
	}
	
	public Element createSchemaElement(String name)
	{
		return fDoc.createElementNS(
						XSCNamespace.SCHEMA_NS,
						XSCNamespace.getSchemaName(name));
	}
	
	public Attr createSchemaAttr(String name)
	{
		return fDoc.createAttribute(name);
	}
	
	public Document getDoc()
	{
		return fDoc;
	}
	
	public XSCSchema getSchema()
	{
		return fSchema;
	}
	
	public boolean isGlobal()
	{
		return fComponents.isEmpty();
	}


	public void startComponent(int type, XSCNameValue name) throws XSCException
	{
		if (fDebug)
		{
			XSCKeywordValue typeKey = new XSCKeywordValue(type);
			String msg = indentString() + "startComponent(" + typeKey.getXMLString() + "," +
						 ((name != null) ? name.getXSCString() : "<null>") + ")";
			System.out.println(msg);
			
			fIndent++;
		}
		
		if (fInComponent)
			throw new XSCException("startComponent() while in component!");
		
		// construct the component
		XSCComponent comp;	
		switch (type)
		{
			case XSCParserConstants.TARGET_NS:
			{ comp = new XSCTargetNamespace(); break; }
			case XSCParserConstants.NAMESPACE:
			{ comp = new XSCNamespace(); break; }
			case XSCParserConstants._DEFAULT:
			{ comp = new XSCDefault(); break; }
			case XSCParserConstants.EL_DEFAULT:
			{ comp = new XSCElementDefault(); break; }
			case XSCParserConstants.ATT_DEFAULT:
			{ comp = new XSCAttributeDefault(); break; }
			case XSCParserConstants.VERSION:
			{ comp = new XSCVersion(); break; }
			case XSCParserConstants.IMPORT:
			{ comp = new XSCImport(); break; }
			case XSCParserConstants.INCLUDE:
			{ comp = new XSCInclude(); break; }
			case XSCParserConstants.REDEFINE:
			{ comp = new XSCRedefine(); break; }
			
			case XSCParserConstants.RESTRICTS:
			{ comp = new XSCRestriction(); break; }
			case XSCParserConstants.LIST:
			{ comp = new XSCList(); break;}
			case XSCParserConstants.UNION:
			{ comp = new XSCUnion(); break;}
			case XSCParserConstants.MIXED:
			{ comp = new XSCContentModel(); break; }
			
			case XSCParserConstants.ELEMENT:
			{ comp = new XSCElement(); break; }
			case XSCParserConstants.ATTRIBUTE:
			{ comp = new XSCAttribute(); break; }
			case XSCParserConstants.COMPLEX_TYPE:
			{ comp = new XSCComplexType(); break; }
			case XSCParserConstants.SIMPLE_TYPE:
			{ comp = new XSCSimpleType(); break; }
			case XSCParserConstants.ANY:
			{ comp = new XSCElementWildcard(); break; }
			case XSCParserConstants.ANY_ATTR:
			{ comp = new XSCAttributeWildcard(); break; }
			case XSCParserConstants.ATT_GROUP:
			{ comp = new XSCAttributeGroup(); break; }
			case XSCParserConstants.NOTATION:
			{ comp = new XSCNotation(); break; }
			
			case XSCParserConstants.OPEN_PAR:
			{ comp = new XSCModelGroup(); break; }
			case XSCParserConstants.GROUP:
			case XSCParserConstants.GROUP_REF:
			{ comp = new XSCGroup(); break; }
			
			case XSCParserConstants.KEY:
			{ comp = new XSCKey(); break; }
			case XSCParserConstants.KEYREF:
			{ comp = new XSCKeyref(); break; }
			case XSCParserConstants.UNIQUE:
			{ comp = new XSCUnique(); break; }
			
			case XSCParserConstants.LENGTH:
			{ comp = new XSCLengthFacet(); break; }
			case XSCParserConstants.OPEN_SQ:
			{ comp = new XSCRangeFacet(); break; }
			case XSCParserConstants.QUOTE_PAT:
			{ comp = new XSCPatternFacet(); break; }
			case XSCParserConstants.SEQ:
			{ comp = new XSCEnumFacet(); break; }
			case XSCParserConstants.WHITE_SPACE:
			{ comp = new XSCWhiteSpaceFacet(); break; }
			case XSCParserConstants.TOTAL_DIG:
			{ comp = new XSCTotalDigitsFacet(); break; }
			case XSCParserConstants.FRACTION_DIG:
			{ comp = new XSCFractionDigitsFacet(); break; }
			
			default:
				throw new XSCException("Unknown component");
		}
		
		// save the current component
		fComponents.push(fCurrent);
		fCurrent = comp;
		fInComponent = true;
		
		// emit qualifiers, the component name and annotations to the new component
		for (int i = 0; i < fQualifiers.size();i++)
			fCurrent.property((XSCKeywordValue)fQualifiers.elementAt(i),0);
			
		for (int i = 0; i < fAnnotations.size();i++)
			fCurrent.property((XSCAnnotationValue)fAnnotations.elementAt(i),0);
		
		if (name != null) fCurrent.property(name,0);	
		
		// delete annotations and qualifiers
		if (fAnnotations.size() > 0) fAnnotations = new Vector();
		if (fQualifiers.size() > 0) fQualifiers = new Vector();
	}

	public void endComponent() throws XSCException
	{
		if (fDebug)
		{
			fIndent--;
			System.out.println(indentString() + "endComponent()");
		}
		
		if (!fInComponent)
			throw new XSCException("endComponent() while not in component");
		
		// fixate the component
		fCurrent.freeze();
		XSCComponent last = fCurrent;
		// switch back to the parent component
		fCurrent = (XSCComponent)fComponents.pop();
		fInComponent = false;
		// emit the ended component as property to its parent component
		fCurrent.property(last,0);
	}

	public void extension(int type, XSCValue value) throws XSCException
	{
		if (fDebug)
		{
			XSCKeywordValue typeKey = new XSCKeywordValue(type);
			String msg = indentString() + "extension(" + typeKey.getXMLString() + "," ;
			
			if (value != null)
				if (value instanceof XSCSimpleValue )
					msg += ((XSCSimpleValue)value).getXSCString();
				else
					msg += value.toString();
			else
				msg += "<null>";
				
			msg += ")";
			
			System.out.println(msg);
		}
		
		if (!fInComponent)
			throw new XSCException("extension() while not in component");
			
		fCurrent.property(value,type);
	}

	public void qualifier(int type) throws XSCException
	{
		XSCKeywordValue keyword = new XSCKeywordValue(type);
		
		if (fDebug)
			System.out.println( indentString() + "qualifier("+keyword.getXMLString()+")");
		
		// when inside component, send directly to component
		// otherwise save for the next component
		if (fInComponent)
			fCurrent.property(keyword,0);
		else
			fQualifiers.add(keyword);
	}

	public void startBlock() throws XSCException
	{
		if (fDebug)
		{
			System.out.println(indentString() + "startBlock()");
			fIndent++;
		}
		
		fInComponent = false;
	
	}

	public void endBlock() throws XSCException
	{
		if (fDebug)
		{
			fIndent--;
			System.out.println(indentString() + "endBlock()");
		}
		
		fInComponent = true;
	}

	public void annotation(String text) throws XSCException
	{
                text = text.indexOf("\n") != -1 ? text : text.trim();
		if (fDebug)
			System.out.println(indentString() + "annotation("+text+")");
		
		// inside schema attach directly
		// else save for next component
		if (!fStripAnnots)
		{
                        if (fCurrent == fSchema)
				fCurrent.property(new XSCAnnotationValue(text),0);
                        else
				fAnnotations.add(new XSCAnnotationValue(text));
		}
	}
	
	private String indentString()
	{
		StringBuffer buf = new StringBuffer();
		for (int i = 0; i < fIndent;i++) buf.append(' ');
		return buf.toString();
	}
}
