package noown.domain.xsc;

import java.io.IOException;

import noown.domain.xsc.val.XSCValue;

/**
 * Compact XML Schema exception class.
 * 
 * @author Kilian Stillhard
 */
public class XSCException extends Exception {
	
	private int fCode;
	private Object fAddition;
	private int fLine;
	private int fCol;
	private boolean fLocDefined;
	
	public static final int Undefined            = 0;
	public static final int ParseException       = 1;
	public static final int NoHandler            = 2;
	
	public static final int DifferentCompositors = 3;
	
	public static final int IllegalProperty      = 4;
	
	public static final int IllQGlobalAttr       = 5;
	public static final int IllQAttr             = 6;
	public static final int IllQAttrDefault      = 7;
	public static final int DupAttWC             = 8;
	public static final int WrongUseQual         = 9;
	public static final int IllQComplexType      = 10;
	public static final int DupSTinCT            = 11;
	public static final int DupCMinCT            = 12;
	public static final int DupTypeDefInCT       = 13;
	public static final int DupElementInCT       = 14;
	public static final int ListUnionInCT        = 15;
	public static final int DupAttWCinCT         = 16;
	public static final int ElNotInCM            = 17;
	public static final int IllQDefault          = 18;
	public static final int IllQElDefault        = 19;
	public static final int FixedEnum            = 20;
	public static final int FixedFracDig         = 21;
	public static final int DupCMinGroup         = 22;
	public static final int DupElInGroup         = 23;
	public static final int FixedLength          = 24;
	public static final int ListOfList           = 25;
	public static final int FixedPattern         = 26;
	public static final int DupTNS               = 27;
	public static final int DupDefault           = 28;
	public static final int DupElDefault         = 29;
	public static final int DupAttDefault        = 30;
	public static final int DupVersion           = 31;
	public static final int IllQSimpleType       = 32;
	public static final int FixedTotDig          = 33;
	public static final int FixedWS              = 34;
	public static final int OtherInNS            = 35;
	public static final int DerAndSTinCT         = 36;
	public static final int TokenMgrErr          = 37;
	public static final int AttrsInLocalAG       = 38;
	public static final int XMLPrefix            = 39;
	public static final int IllIntRangeValue     = 40;
	
	
	private static final String[] fMessages =
	{
		"Undefined error",
		"Parse Exception",
		"No parser event handler defined",
		"Different compositors in model group",
		"Illegal property",
		"Illegal qualifier for global attribute",
		"Illegal qualifier for attribute",
		"Illegal qualifier for attributeDefault",
		"More than one attribute wildcard in attributeGroup",
		"More than one use qualifier",
		"Illegal qualifier for comlexType",
		"More than one simpleType inside complexType",
		"More than one contentModel inside complexType",
		"Both simpleType and contentModel inside complexType",
		"Two Element declarations with same name",
		"List or union simpleType inside complexType",
		"More than one attribute wildcard in complexType",
		"Element not found in content model",
		"Illegal qualifier for default",
		"Illegal qualifier for elementDefault",
		"Cannot declare enumeration facet fixed",
		"Cannot use fixed-minimum or fixed-maximum with fractionDigits",
		"More than one contentModel in group",
		"Two Element declarations with same name inside group",
		"Cannot use fixed-minimum or fixed-maximum with length=number",
		"Cannot use lists of lists",
		"Cannot declare pattern facet fixed",
		"Duplicate targetNamespace declaration",
		"Duplicate default declaration",
		"Duplicate elementDefault declaration",
		"Duplicate attributeDefault declaration",
		"Duplicate version specification",
		"Illegal qualifier for simpleType",
		"Cannot use fixed-minimum or fixed-maximum with totalDigits",
		"Cannot use fixed-minimum or fixed-maximum with whiteSpace",
		"Cannot combine ##other with other namespaces in wildcard namespace specification",
		"Derivation extension in complexType with simpleContent is illegal",
		"Lexer Error",
		"Cannot define local attributeGroups",
		"Cannot use prefixes starting with xml",
		"Illegal value in integer range"
	};

	public XSCException(String message)
	{
		super(message);
	}

	public XSCException(String message, Throwable cause)
	{
		super(message,cause);
	}
	
	public XSCException(int code)
	{
		super(fMessages[code]);
		fCode = code;
	}
	
	
	public XSCException(int code, Object addition)
	{
		super(fMessages[code]);
		fCode = code;
		fAddition = addition;
	}
	
	public XSCException(int code, Throwable cause)
	{
		super(fMessages[code],cause);
		fCode = code;
	}
	
	public void setLocation(int line, int col)
	{
		fLocDefined = true;
		fLine = line;
		fCol = col;
	}
	
	public String getMessage()
	{
		String msg = super.getMessage();
		
		if (fAddition != null)
			msg += " (" + fAddition + ")";
		
		if (fLocDefined)
			msg += "\n" + "At line: " + fLine + " col: " + fCol;
			
		if (getCause() != null)
			msg += "\n" + getCause().getMessage();
		
		return msg;				
	}
	
	public boolean locationDefined()
	{
		return fLocDefined;
	}
	
	public int getLine()
	{
		return fLine;
	}
	
	public int getCol()
	{
		return fCol;
	}
	
	public int getCode()
	{
		return fCode;
	}
	
}
