package noown.domain.xsc.serialize;

import java.io.IOException;
import java.io.Writer;
import java.util.Hashtable;
import java.util.Stack;

import noown.domain.xsc.XSCException;
import noown.domain.xsc.XSCParserConstants;
import noown.domain.xsc.val.XSCKeywordValue;
import noown.domain.xsc.val.XSCNameValue;
import noown.domain.xsc.val.XSCPatternValue;
import noown.domain.xsc.val.XSCSimpleValue;
import noown.domain.xsc.val.XSCStringValue;
import noown.domain.xsc.val.XSCURIValue;
import noown.domain.xsc.val.XSCValue;
import noown.domain.xsc.val.XSCXPathValue;

/**
 * Compact XML Schema Formatter class. 
 * 
 * This class formats the output of XSCSerialize. It acts as a filter
 * between the serializer and the output stream that has to be written to.
 * 
 * @author Kilian Stillhard
 * @version 0.20
 */
public class XSCFormatter
{
	// Internal States
	// Blocks contain one or more component
	// components can contain inner blocks
	// two IN_BLOCK states, one only before the first component was
	// written, a second for the rest.
	//
	// ---------   endBlock()     ---------                  ---------
	// |       | startComponent() |       |  endComponent()  |       |
	// | BLOCK | ---------------> | IN    | ---------------> | BLOCK |
	// |       |                  |       |                  |       |
	// | FIRST | <--------------- | COMP  | <--------------- | SEC   |
	// |       |  startBlock()    |       | startComponent() |       |
	// ---------                  ---------   endBlock()     ---------
	
	// Internal state (see state graph)
	private static final short IN_BLOCK_FIRST = 0;
	private static final short IN_BLOCK_SEC = 1;
	private static final short IN_COMP = 2;
	private short fState;
	
	// Current position on a line
	private int fCol;
	// Current indentation depth
	private int fIndent;
	// Stack to keep up with blocks
	private Stack fBlockStack;
	// Component termination is pending
	private boolean fTermPending;
	// Block start is pending
	private boolean fBlockStartPending;
	// First token on line
	private boolean fHadNewline;
	// Last Token written
	private int fLastTok;
	// First component
	private boolean fFirstComponent;
	
	// The string used for indentation, gets repeated fIndent times
	private String fIndentString;
	// The string used for newlines
	private String fNewLine;
	// Maximum line length
	private int fLineLength;

	// Our writer
	private Writer fWriter;
	
	/**
	 * Create an XSCFormatter with default settings.
	 * 
	 * This creates an XSCFormatter with the settings:
	 * newLine = "LF"
	 * indentString = "\t"
	 * lineLength = 80
	 */ 
	public XSCFormatter(Writer writer)
	{
		this(writer,null,"\t",80);
	}
	
	/**
	 * Create an XSCFormatter.
	 * 
	 * @param writer: A java.io.Writer to which the output is written
	 * @param newLine: Either "CR", "CR-LF" or "LF"
	 * @param indentString: The string used for indentation (i. e. "\t")
	 * @param lineLength: maximum length of a line (or 0 for no wrapping)
	 * */
	public XSCFormatter(Writer writer, 
						 String newLine, 
						 String indentString, 
						 int lineLength)
	{
		fState = IN_BLOCK_FIRST;
		fCol = 0;
		fIndent = 0;
		fTermPending = false;
		fBlockStartPending = false;
		fHadNewline = true;
		fLastTok = 0;
		fFirstComponent = true;
		
		fBlockStack = new Stack();
		fBlockStack.push(new Boolean(false));
		
		fWriter = writer;
		fIndentString = indentString;
		fLineLength = lineLength;
		
		if (newLine != null)
		{	
			if (newLine.equals("CR")) fNewLine = "\r";
			if (newLine.equals("CR-LF")) fNewLine = "\r\n";
			if (newLine.equals("LF")) fNewLine = "\n";
		} else {
			fNewLine = "\n";
		}
	}


	/**
	 * Start a component.
	 * 
	 * This starts a new component. Writing anything is possible only
	 * after this routine has been called. Components are terminated with 
	 * endComponent().
	 * 
	 * @see endComponent();
	 */
	public void startComponent() throws XSCException
	{
		short old_state = fState;
		boolean inline = ((Boolean)fBlockStack.peek()).booleanValue();
		
		// Update state (must be done before writing anything)
		if (fState == IN_BLOCK_FIRST || fState == IN_BLOCK_SEC) fState = IN_COMP;
		else throw new XSCException("Unexpected startComponent() in state IN_COMP!");
		
		try {
			if (old_state == IN_BLOCK_SEC)
			{
				// Check if a prior Component has to be terminated
				// Only terminate components on the same line
				if (fTermPending && inline)
				{
					writeToken(XSCParserConstants.END_COMP);
					fTermPending = false;
					fHadNewline = false;
				}
			
				if (!inline)
				{
					fWriter.write(fNewLine);
					for (int i = 0; i < fIndent; i++) fWriter.write(fIndentString);
					fCol = fIndent*fIndentString.length();
					fHadNewline = true;
				}
			} else {
				if (fBlockStartPending)
				{
					writeToken(XSCParserConstants.START_BLOCK);
					fBlockStartPending = false;
					fHadNewline = false;
				}
				if (!inline && !fFirstComponent)
				{
					fWriter.write(fNewLine);
					for (int i = 0; i < fIndent; i++) fWriter.write(fIndentString);
					fCol = fIndent*fIndentString.length();
					fHadNewline = true;
				}
			}
			fFirstComponent = false;		
		} catch (IOException e)
		{
			throw new XSCException("IO Exception",e);
		}
	}
	
	/**
	 * Ends a component.
	 * 
	 * Must be called at the end of a component. Cannot write anymore
	 * when endComponent() has been called.
	 * 
	 * @see startComponent()
	 */
	public void endComponent() throws XSCException
	{
		fTermPending = true;

		if (fState == IN_COMP) fState = IN_BLOCK_SEC;
		else throw new XSCException("Unexpected endComponent() in state IN_BLOCK!");
	}

	public void startBlock(boolean inline) throws XSCException
	{
		if (fState == IN_COMP) fState = IN_BLOCK_FIRST;
		else throw new XSCException("Unexpected startBlock() in state BLOCK!");	
		
		if (((Boolean)fBlockStack.peek()).booleanValue())
			inline = true;
		
		fBlockStack.push(new Boolean(inline));
		
		if (!inline) fIndent++;
		
		// begin the new block
		fBlockStartPending = true;
	}

	public void endBlock() throws XSCException
	{	
		boolean inline = ((Boolean)fBlockStack.pop()).booleanValue();
		
		// update state (must be done before writing anything)
		if (fState == IN_BLOCK_FIRST || fState == IN_BLOCK_SEC) fState = IN_COMP;
		else throw new XSCException("Unexpected endBlock() in state IN_COMP!");
		
		// terminate components at end of blocks if it is not an inline block
		//if (fTermPending && !inline)
		//	writeToken(XSCParserConstants.END_COMP);
		fTermPending = false;
		
		if (!inline) fIndent--;
		
		// empty block = nothing to do
		if (!fBlockStartPending)
		{
			try {
				if (inline)
				{
					if (!fHadNewline) fWriter.write(" ");
					writeToken(XSCParserConstants.END_BLOCK);
					fCol += 2;
					fHadNewline = false;
				} else {
					fWriter.write(fNewLine);
					for (int i = 0; i < fIndent; i++) fWriter.write(fIndentString);
					writeToken(XSCParserConstants.END_BLOCK);
					fCol = fIndent*fIndentString.length() + 1;
					fHadNewline = false;
				}
			} catch (IOException e)
			{
				throw new XSCException("IO Exception",e);
			}
		} else {
			fBlockStartPending = false;
		}
	}
	
	public void writeToken(int token) throws XSCException
	{
		boolean inline = ((Boolean)fBlockStack.peek()).booleanValue();
		
		if (fState != IN_COMP)
			throw new XSCException("Can writeValue() only when inside component!");
		
		// retrieve image of token, strip quotes
		String tok;
		try {
			tok = XSCParserConstants.tokenImage[token];
		} catch (ArrayIndexOutOfBoundsException e)
		{
			throw new XSCException("Illegal keyword number " + token + "!");
		}
		if (tok.charAt(0) == '"')
			tok = tok.substring(1,tok.length()-1);
			
		try {
			if (fLineLength > 0 && fCol + tok.length()  > fLineLength)
			{
				fWriter.write(fNewLine);
				for (int i = 0; i < fIndent+1; i++) fWriter.write(fIndentString);
				fCol = (fIndent+1)*fIndentString.length();
				fHadNewline = true;
			}
			
			if (!(token == XSCParserConstants.SEQ
			    || token == XSCParserConstants.ONE_OR_MORE
			    || token == XSCParserConstants.ZERO_OR_MORE
			    || token == XSCParserConstants.ONE_OR_ZERO
			    || token == XSCParserConstants.END_COMP
			    || (token == XSCParserConstants.END_BLOCK && !inline)) && !fHadNewline)
			{
				fWriter.write(" ");
				fCol++;
			}
			
			fWriter.write(tok);
			fCol += tok.length();
			fHadNewline = false;
			fLastTok = token;
			
		} catch (IOException e)
		{
			throw new XSCException("IO Exception",e);
		}
	}
	
	public void writeString(String string) throws XSCException
	{
		XSCStringValue str = new XSCStringValue(string,false);
		writeValue(str);
	}
	
	public void writeName(String name) throws XSCException
	{
		XSCNameValue nam = new XSCNameValue(name);
		writeValue(nam);
	}
	
	public void writePattern(String pattern) throws XSCException
	{
		XSCPatternValue pat = new XSCPatternValue(pattern,false);
		writeValue(pat);
	}
	
	public void writeURI(String uri) throws XSCException
	{
		XSCURIValue uriV = new XSCURIValue(uri,false);
		writeValue(uriV);
	}
	
	public void writeXPath(String xpath) throws XSCException
	{
		XSCXPathValue xp = new XSCXPathValue(xpath,false);
		writeValue(xp);
	}
	
	public void writeValue(XSCSimpleValue v) throws XSCException
	{
		if (fState != IN_COMP)
			throw new XSCException("Can writeValue() only when inside component!");
			
		String xscStr = v.getXSCString();
		
		try {
			if (fLineLength > 0 && fCol + xscStr.length() > fLineLength)
			{
				fWriter.write(fNewLine);
				for (int i = 0; i < fIndent+1; i++) fWriter.write(fIndentString);
				fCol = (fIndent+1)*fIndentString.length();
				fHadNewline = true;
			}
			if (!fHadNewline && fLastTok != XSCParserConstants.GROUP_REF) fWriter.write(" ");
			fWriter.write(xscStr);
			fCol += xscStr.length() + 1;
			fHadNewline = false;
		} catch (IOException e)
		{
			throw new XSCException("IO Exception",e);
		}
	}
	
	public void writeAnnotation(String annot) throws XSCException
	{
		try {
			for (int i = 0; i < fIndent; i++) fWriter.write(fIndentString);
			fWriter.write("/* ");
			fWriter.write(annot);
			fWriter.write(" */");
			fWriter.write(fNewLine);
			for (int i = 0; i < fIndent; i++) fWriter.write(fIndentString);
			fCol = fIndent*fIndentString.length();
			fHadNewline = true;
		} catch (IOException e)
		{
			throw new XSCException("IO Exception",e);
		}
	}

        // write a newline but don't update state (used purely for formatting)
        public void writeNewLine() throws XSCException
        {
                try {
                        fWriter.write(fNewLine);
                } catch (IOException e) {
                        throw new XSCException("IO Exception", e);
                }
        }
	
	public void flush() throws XSCException
	{
		try {
			fWriter.flush();
		} catch (IOException e)
		{
			throw new XSCException("IO Exception",e);
		}
	}
}
