package noown.domain.xsc.serialize;

import java.io.BufferedWriter;
import java.io.IOException;
import java.io.OutputStream;
import java.io.UnsupportedEncodingException;
import java.io.Writer;
import java.io.OutputStreamWriter;
import java.io.StringWriter;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.StringTokenizer;
import java.util.Vector;

import noown.domain.xsc.XSCException;
import noown.domain.xsc.XSCParserConstants;
import noown.domain.xsc.XSCWarning;
import noown.domain.xsc.comp.XSCNamespace;
import noown.domain.xsc.comp.XSCTargetNamespace;
import noown.domain.xsc.util.XSCNodeListImpl;
import noown.domain.xsc.val.XSCNameValue;
import noown.domain.xsc.val.XSCStringValue;
import noown.domain.xsc.val.XSCURIValue;

import org.w3c.dom.Attr;
import org.w3c.dom.Comment;
import org.w3c.dom.DOMException;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;

/**
 * Serialize a DOM Tree containing an XML Schema to compact syntax.
 *
 * @author Kilian Stillhard
 */
public class XSCSerialize
{
	
	private XSCFormatter fFormatter;
	
	// Features
	private boolean fWriteAnnotations;
	private String fEncoding;
	private String fNewLine;
	private String fIndentString;
	private int fLineLength;
        private boolean fIncludeBlankLines;
	
	
	
	//constructors
	public XSCSerialize()
	{
		fWriteAnnotations = true;
		fEncoding = "UTF-8";
		fNewLine = "CR";
		fIndentString = "\t";
		fLineLength = 80;
	}
	
	public boolean getWriteAnnotations()
	{
		return fWriteAnnotations;
	}
	
	public void setWriteAnnotations(boolean writeAnnotations)
	{
		fWriteAnnotations = writeAnnotations;
	}

	public String getEncoding()
	{
		return fEncoding;
	}

	public void setEncoding(String encoding)
	{
		fEncoding = encoding;
	}

	public String getNewLine()
	{
		return fNewLine;
	}

	public void setNewLine(String newLine)
	{
		fNewLine = newLine;
	}
	
	public String getIndentString()
	{
		return fIndentString;
	}
	
	public void setIndentString(String indentString)
	{
		fIndentString = indentString;
	}
	
	public int getLineLength()
	{
		return fLineLength;
	}
	
	public void setLineLength(int lineLength)
	{
		fLineLength = lineLength;
	}
	
	public boolean getIncludeBlankLines()
	{
		return fIncludeBlankLines;
	}
	
	public void setIncludeBlankLines(boolean includeBlankLines)
	{
		fIncludeBlankLines = includeBlankLines;
	}

	public void writeNode(OutputStream destination, Node wnode) throws XSCException
	{
		if (destination ==  null || wnode == null)
			throw new XSCException("No destination OutputStream or node defined");
	
		BufferedWriter buf;
	
		try {
			buf = new BufferedWriter(new OutputStreamWriter(destination,fEncoding));
		} catch (UnsupportedEncodingException e)
		{
			throw new XSCException("Unsupported encoding: " + fEncoding,e);
		}
		
		fFormatter = new XSCFormatter(buf,fNewLine,fIndentString,fLineLength);
		
		if (wnode.getNodeType() == Node.DOCUMENT_NODE)
		{
			Document doc = (Document) wnode;
			Element root = doc.getDocumentElement();
			if ((root != null) && isSchemaElement(root,"schema"))
			{
				writeSchema(root);
				fFormatter.flush();
			} else {
				throw new XSCException("No schema root element found!");
			}
		} else {
			throw new XSCException("Cannot write out document fragments!");
		}
	}

	public String writeToString(Node wnode) throws XSCException
	{
		StringWriter sw = new StringWriter();
		
		fFormatter = new XSCFormatter(sw,fNewLine,fIndentString,fLineLength);
		
		if (wnode.getNodeType() == Node.DOCUMENT_NODE)
		{
			Document doc = (Document) wnode;
			Element root = doc.getDocumentElement();
			if ((root != null) && isSchemaElement(root,"schema"))
			{
				writeSchema(root);
				fFormatter.flush();
			} else {
				throw new XSCException("No schema root element found!");
			}
		} else {
			throw new XSCException("Cannot write out document fragments!");
		}
		return sw.toString();
	}



	
	private void writeSchema(Element root) throws XSCException
	{
		XSCNamespace.resetNamespaces();
		
                // write and drop first top-level annotation
		Node curr;
                NodeList nlist = root.getChildNodes();
		for (int i = 0; i < nlist.getLength(); i++)
                {
			curr = nlist.item(i);
			if (curr.getNodeType() == Node.ELEMENT_NODE)
			{
				if (isSchemaElement((Element)curr,null))
				{
					if (curr.getLocalName().equals("annotation"))
					{
						writeSchemaAnnotation((Element)curr);
                                                root.removeChild(curr);
                                                break;
					}
                                }
                        }
                }
		
		// write schema options
		writeSchemaOptions(root);

		// write out children Elements
		nlist = root.getChildNodes();
		for (int i = 0; i < nlist.getLength(); i++)
		{
			curr = nlist.item(i);
			if (curr.getNodeType() == Node.ELEMENT_NODE)
			{
				if (isSchemaElement((Element)curr,null))
				{
					if (curr.getLocalName().equals("annotation"))
					{
						writeSchemaAnnotation((Element)curr);
					}
						
					else if (curr.getLocalName().equals("include"))
					{
						writeInclude((Element)curr);
					}
						
					else if (curr.getLocalName().equals("import"))
					{
						writeImport((Element)curr);
					}
						
					else if (curr.getLocalName().equals("redefine"))
					{
						writeRedefine((Element)curr);
					}
						
					else if (curr.getLocalName().equals("simpleType"))
					{
						writeSimpleType((Element)curr);
					}
						
					else if (curr.getLocalName().equals("complexType"))
					{
						writeComplexType((Element)curr);
					}
						
					else if (curr.getLocalName().equals("group"))
					{
						writeGroup((Element)curr);
					}
						
					else if (curr.getLocalName().equals("attributeGroup"))
					{
						writeAttributeGroup((Element)curr);
					}
						
					else if (curr.getLocalName().equals("element"))
					{
						writeElement((Element)curr);
					}
					
					else if (curr.getLocalName().equals("attribute"))
					{
						writeAttribute((Element)curr);
					}
						
					else if (curr.getLocalName().equals("notation"))
					{
						writeNotation((Element)curr);
					}
						
					else
                                        {
                                                XSCWarning.add("Unknown Element in XML Schema namespace: "
															+ curr.getNodeName());
                                        }
                                        if (fIncludeBlankLines)
                                                fFormatter.writeNewLine();
				} else {    
					XSCWarning.add("Unknown Element from non-XML Schema namespace: "
											+ curr.getNodeName());
				} //if (isSchemaElement(curr,null))
				continue;
			} // if (curr.getNodeType() == Node.ELEMENT_NODE)
			
		} // for (int i = 0; i < nlist.getLength(); i++)
	}

	private void writeNotation(Element not) throws XSCException
	{
		fFormatter.startComponent();
		
		writeAnnotation(not);
		fFormatter.writeToken(XSCParserConstants.NOTATION);
		fFormatter.writeName(not.getAttribute("name"));
		fFormatter.writeToken(XSCParserConstants.PUBLIC);
		fFormatter.writeString(not.getAttribute("public"));
		fFormatter.writeToken(XSCParserConstants.SYSTEM);
		fFormatter.writeURI(not.getAttribute("system"));
		
		fFormatter.endComponent();
	}

	private void writeAttribute(Element attr) throws XSCException
	{
		fFormatter.startComponent();
		
		//check for annotations
		writeAnnotation(attr);
		
		//write out qualifiers
		writeFinalAttribute(attr.getAttribute("final"));
		writeFormAttribute(attr.getAttribute("form"));
		writeUseAttribute(attr.getAttribute("use"));
		
		fFormatter.writeToken(XSCParserConstants.ATTRIBUTE);
		
		// check if reference or declaration
		if (attr.hasAttribute("ref"))
		{
			fFormatter.writeName(attr.getAttribute("ref"));
		} else {
			fFormatter.writeName(attr.getAttribute("name"));
			
			fFormatter.startBlock(true);
			
			//check for type attribute or inner definition
			if (attr.hasAttribute("type"))
			{
				fFormatter.startComponent();
				fFormatter.writeName(attr.getAttribute("type"));
				fFormatter.endComponent();
			} else {
				//look  for inner simpleType element
				Element innerST = (Element) getSchemaChildren(attr,"simpleType").item(0);
				if (innerST != null) writeAnonSimpleType(innerST,false);
			}
			
			fFormatter.endBlock();
		}
		
		// check for fixed/default declaration
		writeFixedDefaultAttr(attr.getAttribute("fixed"),attr.getAttribute("default"));
		
		fFormatter.endComponent();
	}

	private void writeFixedDefaultAttr(String fix, String def) throws XSCException
	{
		if (fix.length() > 0)
		{
			fFormatter.writeToken(XSCParserConstants.EQUAL);
			fFormatter.writeString(fix);
		} else if (def.length() > 0)
		{
			fFormatter.writeToken(XSCParserConstants.SET_DEF);
			fFormatter.writeString(def);
		}
	}
		

	private void writeUseAttribute(String useAttr) throws XSCException
	{
		if (useAttr.indexOf("required") >= 0) 
			fFormatter.writeToken(XSCParserConstants.REQUIRED);
		if (useAttr.indexOf("prohibited") >= 0) 
			fFormatter.writeToken(XSCParserConstants.PROHIBITED);
	}


	private void writeFormAttribute(String formAttr) throws XSCException
	{
		if (formAttr.indexOf("unqualified") >= 0) 
			fFormatter.writeToken(XSCParserConstants.UNQUALIFIED);
		else if (formAttr.indexOf("qualified") >= 0) 
			fFormatter.writeToken(XSCParserConstants.QUALIFIED);
	}


	private void writeElement(Element el) throws XSCException
	{
		fFormatter.startComponent();
		
		// check for annotations
		writeAnnotation(el);
		
		// check for abstract attribute
		if (el.getAttribute("abstract").indexOf("true") >= 0)
			fFormatter.writeToken(XSCParserConstants.ABSTRACT);
			
		// check for nillable attribute
		if (el.getAttribute("nillable").indexOf("true") >= 0)
			fFormatter.writeToken(XSCParserConstants.NILLABLE);
		
		// check for final / block attributes
		writeBlockAttribute(el.getAttribute("block"));
		writeFinalAttribute(el.getAttribute("final"));
		
		// write name
		fFormatter.writeToken(XSCParserConstants.ELEMENT);
		fFormatter.writeName(el.getAttribute("name"));
		
		// check for substitutionGroup Attribute
		if (el.hasAttribute("substitutionGroup"))
		{
			fFormatter.writeToken(XSCParserConstants.SUBSTITUTES);
			fFormatter.writeName(el.getAttribute("substitutionGroup"));
		}
		
		// check for type attribute
		if (el.hasAttribute("type"))
		{	
			fFormatter.startBlock(true);
			fFormatter.startComponent();
			fFormatter.writeName(el.getAttribute("type"));
			fFormatter.endComponent();
			fFormatter.endBlock();
		} else {
			Element st = (Element) getSchemaChildren(el,"simpleType").item(0);
			Element ct = (Element) getSchemaChildren(el,"complexType").item(0);
		
			// check for id constraints
			NodeList unique = getSchemaChildren(el,"unique");
			NodeList key = getSchemaChildren(el,"key");
			NodeList keyref = getSchemaChildren(el,"keyref");
			
			boolean hasIDC = ((unique.getLength() > 0) || (key.getLength() > 0 ) ||
								(keyref.getLength() > 0));
			
			if (st != null)
			{
				fFormatter.startBlock(!hasIDC);
				writeAnonSimpleType(st,false);
				
			} else if (ct != null)
			{
				writeInheritanceDecl(ct);
				
				fFormatter.startBlock(false);
				writeAnonComplexType(ct);
			} else {
				if (hasIDC) fFormatter.startBlock(true);
			}
		
			
			//write id constraints
			for (int i = 0; i < unique.getLength(); i++)
				writeIDC((Element)unique.item(i));
			for (int i = 0; i < key.getLength(); i++)
				writeIDC((Element)key.item(i));
			for (int i = 0; i < keyref.getLength(); i++)
				writeIDC((Element)keyref.item(i));
				
			if ((st != null) || (ct != null) || hasIDC) fFormatter.endBlock();
		}
		fFormatter.endComponent();
	}

	private void writeIDC(Element idc) throws XSCException
	{
		NodeList fields = getSchemaChildren(idc,"field");
		Element sel = (Element) getSchemaChildren(idc,"selector").item(0);
		
		fFormatter.startComponent();
		
		// check for annotations
		writeAnnotation(idc);
		for (int i = 0; i < fields.getLength(); i++)
			writeAnnotation((Element)fields.item(i));
		writeAnnotation(sel);
		
		//write out name
		if (idc.getLocalName().equals("unique"))
			fFormatter.writeToken(XSCParserConstants.UNIQUE);
		else if (idc.getLocalName().equals("key"))
			fFormatter.writeToken(XSCParserConstants.KEY);
		else if (idc.getLocalName().equals("keyref"))
			fFormatter.writeToken(XSCParserConstants.KEYREF);
			
		fFormatter.writeName(idc.getAttribute("name"));
		
		// write out refer attribute
		if (idc.getLocalName().equals("keyref"))
		{
			fFormatter.writeToken(XSCParserConstants.REFERS);
			fFormatter.writeName(idc.getAttribute("refer"));
		}
		
		// write out fields
		fFormatter.writeToken(XSCParserConstants.FIELD);
		for (int i = 0; i < fields.getLength(); i++)
		{
			fFormatter.writeXPath(((Element)fields.item(i)).getAttribute("xpath"));
			if (i < (fields.getLength() - 1))
				fFormatter.writeToken(XSCParserConstants.SEQ);
		}
		
		// write out selector
		if (sel != null)
		{
			fFormatter.writeToken(XSCParserConstants.IN);
			fFormatter.writeXPath(sel.getAttribute("xpath"));
		}
			
		fFormatter.endComponent();
	}


	private void writeAttributeGroup(Element attrGrp) throws XSCException
	{
		fFormatter.startComponent();
		
		// check for annotations
		writeAnnotation(attrGrp);
		
		fFormatter.writeToken(XSCParserConstants.ATT_GROUP);
		
		// check if we are a reference
		if (attrGrp.hasAttribute("ref"))
		{
			fFormatter.writeName(attrGrp.getAttribute("ref"));
		} else {
			// we are a definition
			fFormatter.writeName(attrGrp.getAttribute("name"));
			fFormatter.startBlock(false);
			
			// look for attributes and attributeGroups
			NodeList attrs = getSchemaChildren(attrGrp, "attribute");
			for (int i = 0; i < attrs.getLength(); i++)
				writeAttribute((Element)attrs.item(i));
		
			NodeList moreAttrGrps = getSchemaChildren(attrGrp,"attributeGroup");
			for (int i = 0; i < moreAttrGrps.getLength(); i++)
				writeAttributeGroup((Element)moreAttrGrps.item(i));
			
			// only one <anyAttribute> element is allowed
			Element anyAttr = (Element) getSchemaChildren(attrGrp,"anyAttribute").item(0);
				if (anyAttr != null) writeAnyAttribute(anyAttr);
				
			fFormatter.endBlock();
		}	
		fFormatter.endComponent();			
	}

	private void writeGroup(Element group) throws XSCException
	{
		// group references are not handled here
		
		fFormatter.startComponent();
		
		// check for annotations
		writeAnnotation(group);
		
		fFormatter.writeToken(XSCParserConstants.GROUP);
		fFormatter.writeName(group.getAttribute("name"));
		fFormatter.startBlock(false);
			
		// possible content of <group> is a subset of <complexType>
		writeAnonComplexType(group);

		fFormatter.endBlock();
		fFormatter.endComponent();
	}

	private void writeComplexType(Element ct) throws XSCException
	{
		fFormatter.startComponent();
		
		// check for annotations
		writeAnnotation(ct);
		
		// check for abstract attribute
		if (ct.getAttribute("abstract").indexOf("true") >= 0)
			fFormatter.writeToken(XSCParserConstants.ABSTRACT);
		
		// check for final / block attributes
		writeBlockAttribute(ct.getAttribute("block"));
		writeFinalAttribute(ct.getAttribute("final"));
		
		//write name
		fFormatter.writeToken(XSCParserConstants.COMPLEX_TYPE);
		fFormatter.writeName(ct.getAttribute("name"));
		
		//write Inheritance declaration (restricts/extends)
		writeInheritanceDecl(ct);
		
		//write complex type def body
		fFormatter.startBlock(false);
		writeAnonComplexType(ct);
		fFormatter.endBlock();
		
		fFormatter.endComponent();
	}

	private void writeAnonComplexType(Element ct) throws XSCException
	{
		// the element that holds content model and attributes
		// i.e. either a <complexType>, an <restriction> or
		// <extension> element.
		Element contentDef = null;
		
		// check for content model (simple/complex)
		Element simpleCont = (Element) getSchemaChildren(ct,"simpleContent").item(0);
		Element complCont = (Element) getSchemaChildren(ct,"complexContent").item(0);
		
		// check for simple content
		if (simpleCont != null)
		{
			// check for annotations
			writeAnnotation(simpleCont);
			
			Element ext = (Element) getSchemaChildren(simpleCont,"extension").item(0);
			Element restr = (Element) getSchemaChildren(simpleCont,"restriction").item(0);
			
			if (ext != null) contentDef = ext;
			if (restr != null) contentDef = restr;
				
			if (contentDef != null) writeAnonSimpleType(simpleCont,true);
			
		//check for complex Content
		} else {
			if (complCont != null)
			{
				// check for annotations
				writeAnnotation(complCont);
				
				Element ext = (Element) getSchemaChildren(complCont,"extension").item(0);
				Element restr = (Element) getSchemaChildren(complCont,"restriction").item(0);
			
				if (ext != null) contentDef = ext;
				else if (restr != null) contentDef = restr;
				else contentDef = ct; // should not happen!!
			} else {
				// if no complexContent element is present, attributes
				// and content model are defined directly inside the complexType element.
				contentDef = ct;
			}
		
			// check if a content model has to be written
			Element particle = null;
		
			// find the top particle
			Element seq =    (Element) getSchemaChildren(contentDef,"sequence").item(0);
			Element choice = (Element) getSchemaChildren(contentDef,"choice").item(0);
			Element all =    (Element) getSchemaChildren(contentDef,"all").item(0);
			Element group =  (Element) getSchemaChildren(contentDef,"group").item(0);
			
			if (seq != null) particle = seq;
			else if (choice != null) particle = choice;
			else if (all != null) particle = all;
			else if (group != null) particle = group;
			
			Vector localElms = null;
			
			if (particle != null) 
			{
				localElms = getLocalElementDecls(particle);
				
				fFormatter.startComponent();
				// see if mixed attribute is set
				if ((ct.getAttribute("mixed").indexOf("true") >= 0)
					|| (complCont != null && (complCont.getAttribute("mixed").indexOf("true") >= 0)))
						fFormatter.writeToken(XSCParserConstants.MIXED);	
				writeParticle(particle,localElms);
				
				fFormatter.endComponent();
			} else {
				// empty content model has not to be specified (it could)
			}

			// look for local element declarations
			if (localElms != null) 
			{
				for (int i = 0; i < localElms.size();i++)
					writeElement((Element)localElms.elementAt(i));
			}
		}
		
		// look for attributes and attributeGroups
		NodeList attrs = getSchemaChildren(contentDef, "attribute");
		for (int i = 0; i < attrs.getLength(); i++)
			writeAttribute((Element)attrs.item(i));
		
		NodeList attrGrp = getSchemaChildren(contentDef,"attributeGroup");
		for (int i = 0; i < attrGrp.getLength(); i++)
			writeAttributeGroup((Element)attrGrp.item(i));
			
		// only one <anyAttribute> element is allowed
		Element anyAttr = (Element) getSchemaChildren(contentDef,"anyAttribute").item(0);
		if (anyAttr != null) writeAnyAttribute(anyAttr);
	}

	private void writeAnyAttribute(Element anyAttr) throws XSCException
	{
		fFormatter.startComponent();
		
		// check for annotations
		writeAnnotation(anyAttr);
		
		// look for process attribute
		writeProcessAttribute(anyAttr.getAttribute("process"));
		
		fFormatter.writeToken(XSCParserConstants.ANY_ATTR);
		
		writeNamespaceAttribute(anyAttr.getAttribute("namespace"));
		
		fFormatter.endComponent();
	}

	private Vector getLocalElementDecls(Element particle)
	{
		Vector localElms = new Vector();
		
		// the given particle cannot be an <element> itself
		// therefore scan all children for element decls first
		NodeList elms = getSchemaChildren(particle,"element");
		for (int i = 0; i < elms.getLength(); i++)
			addLocalElementDecl(localElms, (Element)elms.item(i));
		
		// descend recursively into model groups (<all> elements cannot occur here)
		NodeList seqs = getSchemaChildren(particle,"sequence");
		for (int i = 0; i < seqs.getLength(); i++)
		{
			Vector moreElms = getLocalElementDecls((Element)seqs.item(i));
			for (int j = 0; j < moreElms.size(); j++)
				addLocalElementDecl(localElms,(Element)moreElms.elementAt(j));
		}
		
		// the same game for <choice>s
		NodeList choices = getSchemaChildren(particle,"choice");
		for (int i = 0; i < choices.getLength(); i++)
		{
			Vector moreElms = getLocalElementDecls((Element)choices.item(i));
			for (int j = 0; j < moreElms.size(); j++)
				addLocalElementDecl(localElms,(Element)moreElms.elementAt(j));
		}
		
		return localElms;
	}
	
	private void addLocalElementDecl(Vector vec, Element el)
	{
		if (el.hasAttribute("name") && (!el.hasAttribute("type")))
			vec.add(el);
	}



	private void writeInheritanceDecl(Element ct) throws XSCException
	{
		// only restrictions of complex types count for inheritance
		Element complCont = (Element) getSchemaChildren(ct,"complexContent").item(0);
		if (complCont != null)
		{
			Element restr = (Element) getSchemaChildren(complCont,"restriction").item(0);
			Element ext = (Element) getSchemaChildren(complCont,"extension").item(0);
			if (restr != null)
			{
				fFormatter.writeToken(XSCParserConstants.RESTRICTS);
				fFormatter.writeName(restr.getAttribute("base"));
			} else if (ext != null)
			{
				fFormatter.writeToken(XSCParserConstants.EXTENDS);
				fFormatter.writeName(ext.getAttribute("base"));
			}
		}	
	}

	
	private void writeParticle(Element part,Vector localElms) throws XSCException
	{
		// can be called only in component context
		int separator = -1;
		boolean wroteSeparator = false;
		
		// compositors
		if (isSchemaElement(part,"sequence")) separator = XSCParserConstants.SEQ;
		if (isSchemaElement(part,"choice")) separator = XSCParserConstants.CHOICE;
		if (isSchemaElement(part,"all")) separator = XSCParserConstants.ALL;
		
		// we have a compositor
		if (separator != -1)
		{
			fFormatter.writeToken(XSCParserConstants.OPEN_PAR);
			NodeList children = getSchemaChildren(part,"*");
			for (int i = 0; i < children.getLength();i++)
			{
				if (isSchemaElement((Element)children.item(i),"sequence")
					|| isSchemaElement((Element)children.item(i),"choice")
					|| isSchemaElement((Element)children.item(i),"group")
					|| isSchemaElement((Element)children.item(i),"element")
					|| isSchemaElement((Element)children.item(i),"any"))
				{
					writeParticle((Element)children.item(i),localElms);
					// check if there are more particles, if yes write separator
					if (children.item(i+1) != null &&
						(isSchemaElement((Element)children.item(i),"sequence")
						|| isSchemaElement((Element)children.item(i),"choice")
						|| isSchemaElement((Element)children.item(i),"group")
						|| isSchemaElement((Element)children.item(i),"element")
						|| isSchemaElement((Element)children.item(i),"any")))
					{
						fFormatter.writeToken(separator);
						wroteSeparator = true;
					}
				}
			}
			// write a separator at the end to make compositor clear if
			// no separator has been written yet
			if (!wroteSeparator) fFormatter.writeToken(separator);
			
			fFormatter.writeToken(XSCParserConstants.CLOSE_PAR);
		}
		
		// element
		if (isSchemaElement(part,"element"))
		{
			
			if (part.hasAttribute("ref"))
			{
				fFormatter.writeName(part.getAttribute("ref"));
			} else {
				// check if locally declared
				if (localElms.contains(part))
					fFormatter.writeName(part.getAttribute("name"));
				else
				{
					//inline declaration
					if (part.hasAttribute("type"))
					{
						fFormatter.writeName(part.getAttribute("name"));
						fFormatter.startBlock(true);
						fFormatter.startComponent();
						fFormatter.writeName(part.getAttribute("type"));
						fFormatter.endComponent();
						fFormatter.endBlock();
					} else {
						fFormatter.startBlock(true);
						writeElement(part);
						fFormatter.endBlock();
					}
				}
			}
		}
		
		// group
		if (isSchemaElement(part,"group"))
		{
			fFormatter.writeToken(XSCParserConstants.GROUP_REF);
			fFormatter.writeName(part.getAttribute("ref"));
		}
		
		// any
		if (isSchemaElement(part,"any"))
		{
			fFormatter.startBlock(true);
			writeAnyElement(part);
			fFormatter.endBlock();
		}
		
		// write occurence specifier (minOccurs/maxOccurs)
		writeOccurrenceSpec(part.getAttribute("minOccurs"),part.getAttribute("maxOccurs"));
	}

	private void writeOccurrenceSpec(String min, String max) throws XSCException
	{
		int minOcc = 1;
		int maxOcc = 1;
		
		try {
			if (min.length() > 0) minOcc = (new Integer(min)).intValue();
			if (max.length() > 0) maxOcc = (new Integer(max)).intValue();
		} catch (NumberFormatException e) 
		{
			if (max.indexOf("unbounded") >= 0) maxOcc = -1;
		}
		
		if ((minOcc == 1) && (maxOcc == 1)) return;
		
		if ((minOcc == 0) && (maxOcc == 1)) 
			{ fFormatter.writeToken(XSCParserConstants.ONE_OR_ZERO); return;}
		if ((minOcc == 0) && (maxOcc == -1)) 
			{ fFormatter.writeToken(XSCParserConstants.ZERO_OR_MORE); return; }
		if ((minOcc == 1) && (maxOcc == -1)) 
			{ fFormatter.writeToken(XSCParserConstants.ONE_OR_MORE); return; }
		
		fFormatter.writeToken(XSCParserConstants.OPEN_SQ);
		fFormatter.writeName((new Integer(minOcc)).toString());
		fFormatter.writeToken(XSCParserConstants.SEQ);
		if (maxOcc != -1) 
			fFormatter.writeName((new Integer(maxOcc)).toString());
		fFormatter.writeToken(XSCParserConstants.CLOSE_SQ); 
	}


	private void writeAnyElement(Element any) throws XSCException
	{
		fFormatter.startComponent();
		
		// look for process attribute
		writeProcessAttribute(any.getAttribute("process"));
		
		fFormatter.writeToken(XSCParserConstants.ANY);
		
		writeNamespaceAttribute(any.getAttribute("namespace"));
		
		fFormatter.endComponent();
	}
	
	private void writeNamespaceAttribute(String ns) throws XSCException
	{
		if (ns != null && ns.length() > 0)
		{
			if (ns.indexOf("##any") >= 0) return;
			
			StringTokenizer tok = new StringTokenizer(ns);
			
			if (tok.countTokens() > 0)
				fFormatter.writeToken(XSCParserConstants.NAMESPACE);
			
			while (tok.hasMoreTokens())
			{
				String nsDecl = tok.nextToken();
				if (nsDecl.equals("##other"))
					fFormatter.writeToken(XSCParserConstants.OTHER_CONST);
				else if (nsDecl.equals("##targetNamespace"))
					fFormatter.writeToken(XSCParserConstants.TARGET_NS_CONST);
				else if (nsDecl.equals("##local"))
					fFormatter.writeToken(XSCParserConstants.LOCAL_CONST);
				else
					fFormatter.writeURI(nsDecl);
				if (tok.hasMoreTokens())
					fFormatter.writeToken(XSCParserConstants.SEQ);
			}
		}

	}
	
	private void writeProcessAttribute(String pr) throws XSCException
	{
		if (pr != null && pr.length() > 0)
		{
			if (pr.indexOf("lax") >= 0)
				fFormatter.writeToken(XSCParserConstants.LAX);
			if (pr.indexOf("strict") >= 0)
				fFormatter.writeToken(XSCParserConstants.STRICT);
			if (pr.indexOf("skip") >= 0)
				fFormatter.writeToken(XSCParserConstants._SKIP);
		}
	}


	private void writeSimpleType(Element el) throws XSCException
	{
		fFormatter.startComponent();
		
		//check for annotations
		writeAnnotation(el);
		
		//write out qualifiers
		writeFinalAttribute(el.getAttribute("final"));

		//write out name
		fFormatter.writeToken(XSCParserConstants.SIMPLE_TYPE);
		fFormatter.writeName(el.getAttribute("name"));
		
		//write anonymous simple type definition
		fFormatter.startBlock(true);
		writeAnonSimpleType(el,false);
		fFormatter.endBlock();
		
		fFormatter.endComponent();
	}

	private void writeFinalAttribute(String finalAttr) throws XSCException
	{
		if (finalAttr.indexOf("#all") >= 0) 
			fFormatter.writeToken(XSCParserConstants.FINAL);
		if (finalAttr.indexOf("list") >= 0) 
			fFormatter.writeToken(XSCParserConstants.FINAL_LIST);
		if (finalAttr.indexOf("union") >= 0) 
			fFormatter.writeToken(XSCParserConstants.FINAL_UNION);
		if (finalAttr.indexOf("restriction") >= 0) 
			fFormatter.writeToken(XSCParserConstants.FINAL_RESTR);
		if (finalAttr.indexOf("extension") >= 0) 
			fFormatter.writeToken(XSCParserConstants.FINAL_EXT);
	}
	
	private void writeBlockAttribute(String blockAttr) throws XSCException
	{
		if (blockAttr.indexOf("#all") >= 0) 
			fFormatter.writeToken(XSCParserConstants.BLOCK);
		if (blockAttr.indexOf("substitution") >= 0) 
			fFormatter.writeToken(XSCParserConstants.BLOCK_SUBST);
		if (blockAttr.indexOf("restriction") >= 0) 
			fFormatter.writeToken(XSCParserConstants.BLOCK_RESTR);
		if (blockAttr.indexOf("extension") >= 0) 
			fFormatter.writeToken(XSCParserConstants.BLOCK_EXT);
	}


	private void writeAnonSimpleType(Element el,boolean inCT) throws XSCException
	{
		fFormatter.startComponent();
		
		//check for a restriction
		Element restr = (Element) getSchemaChildren(el,"restriction").item(0);
		if (restr != null)
		{
			// see if there is an inner simpleType definition
			Element innerST = (Element) getSchemaChildren(restr,"simpleType").item(0);
			if (innerST != null)
			{
				//write out inner simpleType definition
				fFormatter.writeToken(XSCParserConstants.SIMPLE_TYPE);
				fFormatter.startBlock(true);
				writeAnonSimpleType(innerST,false);
				fFormatter.endBlock();
			} else {
				//write out base name
				fFormatter.writeName(restr.getAttribute("base"));
			}
			
			// write out facets
			writeFacets(restr, inCT);
		}
		
		//check for an extension
		Element ext = (Element) getSchemaChildren(el,"extension").item(0);
		if (ext != null)
		{
			fFormatter.writeName(ext.getAttribute("base"));
		}
		
		//check for an union
		Element union = (Element) getSchemaChildren(el, "union").item(0);
		if (union != null)
		{
			// start the union
			fFormatter.writeToken(XSCParserConstants.UNION);
			fFormatter.startBlock(false);
			
			// write out memberTypes attribute
			StringTokenizer tok = new StringTokenizer(union.getAttribute("memberTypes"));
			boolean hadMembers = false;
			while (tok.hasMoreTokens())
			{
				hadMembers = true;
				fFormatter.startComponent();
				fFormatter.writeName(tok.nextToken());
				fFormatter.endComponent();
			}
			
			// write out inner simpleType declarations
			NodeList innerST = getSchemaChildren(union,"simpleType");
			for (int i = 0; i < innerST.getLength();i++)
				writeAnonSimpleType((Element)innerST.item(i),false);
			
			fFormatter.endBlock();
		}
		
		//check for a list
		Element list = (Element) getSchemaChildren(el, "list").item(0);
		if (list != null)
		{
			// start the list
			fFormatter.writeToken(XSCParserConstants.LIST);
			fFormatter.startBlock(true);
			
			// check for an itemType attribute
			if (list.getAttribute("itemType").length() > 0)
			{
				fFormatter.startComponent();
				fFormatter.writeName(list.getAttribute("itemType"));
				fFormatter.endComponent();
			} else { // or an inner simpleType definition
				Element innerST = (Element) getSchemaChildren(list,"simpleType").item(0);
				if (innerST != null) writeAnonSimpleType(innerST,false);
			}
			
			fFormatter.endBlock();
		}
		
		fFormatter.endComponent();
	}

	// writes out facets, context node must be a restriction
	private void writeFacets(Element restr,boolean inCT) throws XSCException
	{
		// check if there are any facets
		// if we are inside a complex type, write curly brackets anyway
		// to make clear that this is a restriction
		NodeList facets = getSchemaChildren(restr,"*");
		if (facets.getLength() > 0 || inCT)
		{
			// starts facets section
			fFormatter.startBlock(true);
			
			// look for range facets (min|max)(Inclusive|Exclusive)
			Element minIncl = (Element) getSchemaChildren(restr,"minInclusive").item(0);
			Element minExcl = (Element) getSchemaChildren(restr,"minExclusive").item(0);
			Element maxIncl = (Element) getSchemaChildren(restr,"maxInclusive").item(0);
			Element maxExcl = (Element) getSchemaChildren(restr,"maxExclusive").item(0);

			if ((minExcl != null) || (minIncl != null) || (maxExcl != null) || (maxIncl != null))
				writeRangeFacet(minExcl,minIncl,maxExcl,maxIncl);
			
			// look for length facets (min|max)length
			Element minLength = (Element) getSchemaChildren(restr,"minLength").item(0);
			Element maxLength = (Element) getSchemaChildren(restr,"maxLength").item(0);
			Element Length = (Element) getSchemaChildren(restr,"length").item(0);
			
			if ((minLength != null) || (maxLength != null) || (Length != null))
				writeLengthFacet(minLength,maxLength,Length);
			
			// look for pattern facets
			Element pattern = (Element) getSchemaChildren(restr,"pattern").item(0);
			
			if (pattern != null)
				writePatternFacet(pattern);
			
			// look for enumeration facets
			NodeList enums = getSchemaChildren(restr,"enumeration");
			
			if (enums.getLength() > 0)
				writeEnumerationFacet(enums);
			
			// look for whitspace facet
			Element whiteSp = (Element) getSchemaChildren(restr,"whiteSpace").item(0);
			
			if (whiteSp != null)
				writeWhiteSpaceFacet(whiteSp);
			
			// look for totalDigits facet
			Element totalDig = (Element) getSchemaChildren(restr,"totalDigits").item(0);
			
			if (totalDig != null)
				writeTotalDigitsFacet(totalDig);
			
			// look for fractionDigits facet
			Element fractionDig = (Element) getSchemaChildren(restr,"fractionDigits").item(0);
			
			if (fractionDig != null)
				writeFractionDigitsFacet(fractionDig);
				
			// end facets section
			fFormatter.endBlock();
		}
	}

	private void writeFractionDigitsFacet(Element fractionDig) throws XSCException
	{
		fFormatter.startComponent();
		fFormatter.writeToken(XSCParserConstants.FRACTION_DIG);
		fFormatter.writeToken(XSCParserConstants.EQUAL);
		fFormatter.writeName(fractionDig.getAttribute("value"));
		fFormatter.endComponent();
	}


	private void writeTotalDigitsFacet(Element totalDig) throws XSCException
	{
		fFormatter.startComponent();
		fFormatter.writeToken(XSCParserConstants.TOTAL_DIG);
		fFormatter.writeToken(XSCParserConstants.EQUAL);
		fFormatter.writeName(totalDig.getAttribute("value"));
		fFormatter.endComponent();
	}


	private void writeWhiteSpaceFacet(Element whiteSp) throws XSCException
	{
		fFormatter.startComponent();
		fFormatter.writeToken(XSCParserConstants.WHITE_SPACE);
		fFormatter.writeToken(XSCParserConstants.EQUAL);
		
		String wsAttr = whiteSp.getAttribute("value").trim();
		if (wsAttr.equals("preserve"))
			fFormatter.writeToken(XSCParserConstants.PRESERVE);
		if (wsAttr.equals("collapse"))
			fFormatter.writeToken(XSCParserConstants.COLLAPSE);
		if (wsAttr.equals("replace"))
			fFormatter.writeToken(XSCParserConstants.REPLACE);
		fFormatter.endComponent();
	}


	private void writeEnumerationFacet(NodeList enums) throws XSCException
	{
		fFormatter.startComponent();
		for (int i = 0; i < enums.getLength();i++)
		{
			fFormatter.writeString(((Element)enums.item(i)).getAttribute("value"));
			if (i < (enums.getLength()-1)) 
				fFormatter.writeToken(XSCParserConstants.SEQ);
		}
		fFormatter.endComponent();
	}


	private void writePatternFacet(Element pattern) throws XSCException
	{
		fFormatter.startComponent();
		fFormatter.writePattern(pattern.getAttribute("value"));
		fFormatter.endComponent();
	}


	private void writeLengthFacet(Element minLength,Element maxLength,Element Length) 
						throws XSCException
	{
		fFormatter.startComponent();
		// start length
		fFormatter.writeToken(XSCParserConstants.LENGTH);
		fFormatter.writeToken(XSCParserConstants.EQUAL);
		
		// write length or length range
		if (Length != null)
		{
			fFormatter.writeName(Length.getAttribute("value"));
		} else {
			fFormatter.writeToken(XSCParserConstants.OPEN_SQ);
			if (minLength != null) 
				fFormatter.writeName(minLength.getAttribute("value"));
			fFormatter.writeToken(XSCParserConstants.SEQ);
			if (maxLength != null) 
				fFormatter.writeName(maxLength.getAttribute("value"));
			fFormatter.writeToken(XSCParserConstants.CLOSE_SQ);
		}
		fFormatter.endComponent();
	}


	private void writeRangeFacet(Element minExcl,Element minIncl,
									Element maxExcl,Element maxIncl) throws XSCException
	{
		fFormatter.startComponent();
		
		// start range
		if (minExcl != null) 
			fFormatter.writeToken(XSCParserConstants.OPEN_PAR);
		else 
			fFormatter.writeToken(XSCParserConstants.OPEN_SQ);
				
		// write out range start
		if (minExcl != null) 
			fFormatter.writeName(minExcl.getAttribute("value"));
		else if (minIncl != null) 
			fFormatter.writeName(minIncl.getAttribute("value"));
				
		fFormatter.writeToken(XSCParserConstants.SEQ);
				
		// write out range end
		if (maxExcl != null) 
			fFormatter.writeName(maxExcl.getAttribute("value"));
		else if (maxIncl != null) 
			fFormatter.writeName(maxIncl.getAttribute("value"));
				
		// end range
		if (maxExcl != null) 
			fFormatter.writeToken(XSCParserConstants.CLOSE_PAR);
		else 
			fFormatter.writeToken(XSCParserConstants.CLOSE_SQ);
			
		fFormatter.endComponent();
	}




	private void writeRedefine(Element redef) throws XSCException
	{
		fFormatter.startComponent();
		fFormatter.writeToken(XSCParserConstants.REDEFINE);
		fFormatter.writeString(redef.getAttribute("schemaLocation"));
		
		// check for contents
		NodeList children = getSchemaChildren(redef,"*");
		if (children.getLength() > 0)
		{
			fFormatter.startBlock(false);
			for (int i = 0; i < children.getLength();i++)
			{
				if (isSchemaElement((Element)children.item(i),"simpleType"))
					writeSimpleType((Element)children.item(i));
				if (isSchemaElement((Element)children.item(i),"complexType"))
					writeComplexType((Element)children.item(i));
				if (isSchemaElement((Element)children.item(i),"group"))
					writeGroup((Element)children.item(i));
				if (isSchemaElement((Element)children.item(i),"attributeGroup"))
					writeAttributeGroup((Element)children.item(i));
			}
			fFormatter.endBlock();
		}
		fFormatter.endComponent();
	}

	private void writeImport(Element imp) throws XSCException
	{
		fFormatter.startComponent();
		fFormatter.writeToken(XSCParserConstants.IMPORT);
		fFormatter.writeString(imp.getAttribute("schemaLocation"));
		if (imp.hasAttribute("namespace"))
		{
			fFormatter.writeToken(XSCParserConstants.NAMESPACE);
			fFormatter.writeString(imp.getAttribute("namespace"));
		}
		fFormatter.endComponent();
	}

	private void writeInclude(Element inc) throws XSCException
	{
		fFormatter.startComponent();
		fFormatter.writeToken(XSCParserConstants.INCLUDE);
		fFormatter.writeString(inc.getAttribute("schemaLocation"));
		fFormatter.endComponent();
	}

	private void writeAnnotation(Element el) throws XSCException
	{
		// looks for <annotation> elements for the child elements
		if (el == null) return;
		Element annot = (Element) getSchemaChildren(el,"annotation").item(0);
		
		if (annot != null)
			writeSchemaAnnotation(annot);
	}
	
	private void writeSchemaAnnotation(Element annot) throws XSCException
	{
		if (fWriteAnnotations)
		{
			// we only consider xs:documentation
			StringBuffer buf = new StringBuffer();
			String ls = System.getProperty("line.separator");
				
			NodeList comments = getSchemaChildren(annot,"documentation");
			for (int i = 0; i < comments.getLength();i++)
			{
				Element docu = (Element)comments.item(i);
				NodeList texts = docu.getChildNodes();
				if (i > 0) buf.append(ls);
				for (int j = 0;j < texts.getLength();j++)
				{
					if (texts.item(j).getNodeType() == Node.TEXT_NODE)
                                            buf.append(texts.item(j).getNodeValue());
				}
                                // trim trailing whitespace, but not newlines
                                int pos = buf.length() - 1;
                                while (Character.isWhitespace(buf.charAt(pos))
                                    && (Character.getType(buf.charAt(pos)) !=
                                        Character.LINE_SEPARATOR)
                                       && (buf.charAt(pos) != '\n')
                                       && (buf.charAt(pos) != '\r'))
                                        pos--;
                                buf.delete(pos+1, buf.length());
			}
			if (buf.length() > 0) {
                            if (!fIncludeBlankLines)
                                    fFormatter.writeNewLine();
                            fFormatter.writeAnnotation(buf.toString());
                        }
		}
	}

	private void writeSchemaOptions(Element root) throws XSCException
	{
		// check for targetNamespace attribute
		if (root.hasAttribute("targetNamespace"))
		{
			fFormatter.startComponent();
			fFormatter.writeToken(XSCParserConstants.TARGET_NS);
			
			String tns = root.getAttribute("targetNamespace");
			fFormatter.writeValue(new XSCURIValue(tns,false));
			
			fFormatter.endComponent();
			
			XSCTargetNamespace.setTargetNamespace(tns);
		}
		
		// check for namespace declarations
		NamedNodeMap attrs = root.getAttributes();
		for (int i = 0; i < attrs.getLength(); i++)
		{
			Attr at = (Attr)attrs.item(i);
			if (at.getPrefix() != null && at.getPrefix().equals("xmlns"))
			{
				if (!(at.getLocalName().equals("xs") && at.getValue().equals(XSCNamespace.SCHEMA_NS)))
				{
					fFormatter.startComponent();
					fFormatter.writeToken(XSCParserConstants.NAMESPACE);
					fFormatter.writeName(at.getLocalName());
					fFormatter.writeURI(at.getValue());
					fFormatter.endComponent();
				}
				XSCNamespace.addNamespace(at.getLocalName(),at.getValue());
			}
			else if (at.getLocalName().equals("xmlns") && at.getPrefix() == null)
			{
				XSCNamespace.addNamespace("",at.getValue());
				fFormatter.startComponent();
				fFormatter.writeToken(XSCParserConstants.NAMESPACE);
				fFormatter.writeURI(at.getValue());
				fFormatter.endComponent();
			}
		}
		
		// check for elementFormDefault attribute (only when "unqualified")
		if (root.hasAttribute("elementFormDefault") 
				&& (root.getAttribute("elementFormDefault").indexOf("unqualified") >= 0))
		{
			fFormatter.startComponent();
			fFormatter.writeToken(XSCParserConstants.EL_DEFAULT);
			fFormatter.writeToken(XSCParserConstants.UNQUALIFIED);
			fFormatter.endComponent();
		}
		
		// check for attributeFormDefault attribute (only when "qualified")
		if (root.hasAttribute("attributeFormDefault") 
				&& (root.getAttribute("attributeFormDefault").indexOf("unqualified") < 0)
				&& (root.getAttribute("attributeFormDefault").indexOf("qualified") >= 0))
		{
			fFormatter.startComponent();
			fFormatter.writeToken(XSCParserConstants.ATT_DEFAULT);
			fFormatter.writeToken(XSCParserConstants.QUALIFIED);
			fFormatter.endComponent();
		}
		
		// check for blockDefault and finalDefault attributes
		if (root.hasAttribute("blockDefault") || root.hasAttribute("finalDefault"))
		{
			fFormatter.startComponent();
			fFormatter.writeToken(XSCParserConstants._DEFAULT);
			writeBlockAttribute(root.getAttribute("blockDefault"));
			writeFinalAttribute(root.getAttribute("finalDefault"));
			fFormatter.endComponent();
		}
		
		// check for version attribute
		if (root.hasAttribute("version"))
		{
			fFormatter.startComponent();
			fFormatter.writeToken(XSCParserConstants.VERSION);
			fFormatter.writeString(root.getAttribute("version"));
			fFormatter.endComponent();
		}
	}

	private boolean isSchemaElement(Element el, String name)
	{
		if (el.getNamespaceURI().equals(XSCNamespace.SCHEMA_NS) 
			&& ( name == null || el.getLocalName().equals(name)))
		{
			return true;
		} else {
			return false;
		}
	}
	
	private NodeList getSchemaChildren(Element el, String name)
	{
		XSCNodeListImpl list = new XSCNodeListImpl();
		
		Node curr= el.getFirstChild();
		
		while (curr != null)
		{
			if (curr.getNodeType() == Node.ELEMENT_NODE
					&& curr.getNamespaceURI().equals(XSCNamespace.SCHEMA_NS)
					&& (name.equals("*") || curr.getLocalName().equals(name)))
			{
				list.add(curr); 
			}
			curr = curr.getNextSibling();
		}
		
		return list;		
	}
}
