package noown.domain.xsc.util;

import java.io.File;
import java.io.FileOutputStream;
import java.io.ByteArrayOutputStream;
import java.io.BufferedOutputStream;
import java.io.OutputStream;
import java.io.IOException;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;

import noown.domain.xsc.XSCException;
import noown.domain.xsc.XSCWarning;
import noown.domain.xsc.serialize.XSCSerialize;
import org.w3c.dom.Document;
import org.xml.sax.SAXException;


/**
 * Main Class for the Conversion from XML to compact syntax.
 * 
 * @author Kilian Stillhard
 */
public class XMLtoXSC
{
        /** Constant for newline separator specification: Linefeed. */
        public static final int NEWLINE_LF = 1;
        /** Constant for newline separator specification: Carriage-return. */
        public static final int NEWLINE_CR = 1;
        /** Constant for newline separator specification:
            CarriageReturn-Linefeed. */
        public static final int NEWLINE_CR_LF = 1;
        /** Constant for newline separator specification: Use system default. */
        public static final int NEWLINE_DEFAULT = -1;

        // settings
        private boolean skipAnnotations = false;
        private String indentString = "   ";
        private String newLine;
        private String encoding = null;
        private int lineLength = 0;
        private boolean validate = false;
        private boolean includeBlankLines = true;

        // these are used if user specified file input/output
        String inFile = null;
        String outFile = null;


        private XMLtoXSC()
        {
		String sep = System.getProperty("line.separator");
		if (sep.equals("\n")) newLine = "LF";
		else if (sep.equals("\r\n")) newLine = "CR-LF";
		else if (sep.equals("\r")) newLine = "CR";
		else newLine = "LF";
		XSCWarning.resetWarnings();
        }

        private XMLtoXSC(boolean skipAnnotations,
                         boolean validateInputFile,
                         int indentation,
                         int newline,
                         int lineLength,
                         boolean includeBlank,
                         String encoding)
        {
                this.skipAnnotations = skipAnnotations;
                this.validate = validateInputFile;

                if (indentation == -1)
                    this.indentString = "\t";
                else if (indentation == 0)
                    this.indentString = "";
                else {
                    indentString = "";
                    for (int i=0; i<indentation; i++)
                        this.indentString += " ";
                }

                if (newline == NEWLINE_LF) this.newLine = "LF";
                else if (newline == NEWLINE_CR) this.newLine = "CR";
                else if (newline == NEWLINE_CR_LF) this.newLine = "CR-LF";

                this.encoding = encoding;
                this.lineLength = lineLength;

                this.includeBlankLines = includeBlank;
        }

        // called only from main(); OK to System.exit()
        private void processArguments(String[] args)
        {
		for (int i = 0; i < args.length; i++)
		{
			if (args[i].equals("-v"))
			{ validate = true; continue; }
			if (args[i].equals("-a"))
			{ skipAnnotations = true; continue; }
			
			if (args[i].equals("-i"))
			{ 
				int n = 0;
				StringBuffer buf = new StringBuffer();
				
				if (i + 1 >= args.length)
				{ usage(); System.exit(1); }
				
				if (args[++i].trim().equals("t"))
				{ indentString = "\t"; continue; }
				
				try { n = Integer.parseInt(args[i]); }
				catch (NumberFormatException e) { usage(); System.exit(1); }
				
				if (n < 1) { usage(); System.exit(1); }
				
				for (int j = 0; j < n;j++) { buf.append(" "); }
				indentString = buf.toString();
				continue;
			}
			
			if (args[i].equals("-n"))
			{
				if (i + 1 >= args.length)
				{ usage(); System.exit(1); }
				
				i++;
				if (args[i].equals("LF"))    newLine = "LF";
				else if (args[i].equals("CR"))    newLine = "CR";
				else if (args[i].equals("CR-LF")) newLine = "CR-LF";
				else { usage(); System.exit(1); }
				continue;
			}
			
			if (args[i].equals("-e"))
			{
				if (i + 1 >= args.length)
				{ usage(); System.exit(1); }
				
				encoding = args[++i].trim();
				continue;
			}
			
			if (args[i].equals("-w"))
			{
				if (i + 1 >= args.length)
				{ usage(); System.exit(1); }
				
				int n = 0;
				try { n = Integer.parseInt(args[++i]); }
				catch (NumberFormatException e) { usage(); System.exit(1); }
				
				if (n < 20) { usage(); System.exit(1); }
				lineLength = n;
				continue;
			}
			
			if (args[i].equals("-c"))
			{
                                includeBlankLines = false;
				continue;
			}
			
			if (inFile == null) inFile = args[i];
			else if (outFile == null) outFile = args[i];
			else { usage(); System.exit(1); }
		}
		
		if (inFile == null) { usage(); System.exit(1); }
		if (outFile == null)
                        outFile = addExtension(inFile);
        }


        private Document parseInputFile(String inFile)
            throws SAXException, IOException
        {
		DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
		factory.setValidating(validate);
		factory.setNamespaceAware(true);
		
		DocumentBuilder parser = null;
		try {
			parser = factory.newDocumentBuilder();
		} catch (Exception e) { e.printStackTrace(); System.exit(1); }
		
		return parser.parse(inFile);
        }


        private void writeOutput(Document inDoc, OutputStream out)
        {
		XSCSerialize converter = new XSCSerialize();
		converter.setIndentString(indentString);
		
		if (encoding != null) converter.setEncoding(encoding);
		converter.setLineLength(lineLength);
		converter.setNewLine(newLine);
		converter.setWriteAnnotations(!skipAnnotations);

                converter.setIncludeBlankLines(includeBlankLines);
		
		try { converter.writeNode(out, inDoc); }
		catch (XSCException e) { System.err.println(e.getMessage()); }
        }


        /**
         *  API method: DOM input, String output; takes arguments directly.
         *  See {@link #convert(boolean,boolean,int,int,String,int)} for
         *  information on default settings.
         */
        public static String convert(Document doc)
            throws SAXException,IOException

        {
            return convert(doc, false, false, 3, NEWLINE_DEFAULT, 0,true, null);
        }

        /**
         *  API method: DOM input, String output; takes arguments directly.
         *  
         *  @param skipAnnotations  Don't convert annotations (false)
         *  @param validateInputFile report whether schema well-formed (false)
         *  @param indentation  indent unit (#/spaces); 0=none, -1=use tab (3)
         *  @param newline  {NEWLINE_LF,NEWLINE_CR,NEWLINE_CR_LF} for Unix,
         *                  Mac, or Windows; special value NEWLINE_DEFAULT will
         *                  select platform default (NEWLINE_DEFAULT)
         *  @param encoding specify encoding to use, null for platform default
         *                  (null)
         *  @param lineLength wrap lines after this many characters,
         *                    0 means don't wrap (0)
         *  @param includeBlank include blank lines in output for readability
         *                      (true)
         */
        public static String convert(Document doc,
                                     boolean skipAnnotations,
                                     boolean validateInputFile,
                                     int indentation,
                                     int newline,
                                     int lineLength,
                                     boolean includeBlank,
                                     String encoding)
            throws SAXException,IOException
        {
                // init
                XMLtoXSC converter = new XMLtoXSC
                    (skipAnnotations, validateInputFile, indentation, newline,
                     lineLength, includeBlank, encoding);

                // go
                ByteArrayOutputStream ba = new ByteArrayOutputStream();
                BufferedOutputStream bu = new BufferedOutputStream(ba);
                converter.writeOutput(doc, bu);
                bu.close();
                return ba.toString();
        }


        /**
         *  Command-line method: file input/output; reads arguments as switches.
         */
	public static void main(String[] args) throws SAXException,IOException
	{
                XMLtoXSC converter = new XMLtoXSC();
                converter.processArguments(args);

                Document doc = converter.parseInputFile(converter.inFile);
		
		FileOutputStream out = new FileOutputStream(converter.outFile);

                converter.writeOutput(doc, out);
		
		out.close();
		
		XSCWarning.printWarnings(System.err);
	}

	
	private static String addExtension(String infile)
	{
		// find last dot
		String namepart = null;
		
		int dotIndex = infile.lastIndexOf('.');
		if (dotIndex <= 0) namepart = infile;
		else namepart = infile.substring(0,dotIndex);
		
		// check if exists
		File f = new File(namepart + ".xsc");
		if (!f.exists()) return (namepart + ".xsc");
		
		// create another filename
		String newNamepart = null;
		for (int i = 1;true;i++)
		{
			f = new File(namepart + i + ".xsc");
			if (!f.exists()) return (namepart + i + ".xsc");
		}
	}

	
	private static void usage()
	{
		System.out.println("usage: java noown.domain.xsc.util.XMLtoXSC [options] xmlfile compactfile");
		System.out.println("options: -a : skip annotations");
		System.out.println("         -v : validate input file");
		System.out.println("         -i number or t: use specified number of blanks (or tab) for indentation (default 3)");
		System.out.println("         -n newline: specify newline (CR, CR-LF or LF)");
		System.out.println("         -w number: wrap lines after number characters");
		System.out.println("         -c: compact format: do not insert blank lines between definitions");
		System.out.println("         -e encoding: specify encoding to use");
	}
}
