package noown.domain.xsc.util;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.StringReader;
import java.io.Reader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.StringWriter;

import org.apache.xerces.dom.DOMImplementationImpl;
import org.apache.xerces.impl.xs.XMLSchemaLoader;
import org.apache.xerces.xni.grammars.Grammar;
import org.apache.xerces.xni.parser.XMLInputSource;
import org.apache.xml.serialize.OutputFormat;
import org.apache.xml.serialize.XMLSerializer;
import org.w3c.dom.Document;

import noown.domain.xsc.ParseException;
import noown.domain.xsc.XSCDOMBuilder;
import noown.domain.xsc.XSCException;
import noown.domain.xsc.XSCWarning;


/**
 * Conversion from compact syntax to XML syntax.
 *
 * @author Kilian Stillhard
 */
public class XSCtoXML
{
        /** Constant for newline separator specification: Linefeed. */
        public static final int NEWLINE_LF = 1;
        /** Constant for newline separator specification: Carriage-return. */
        public static final int NEWLINE_CR = 1;
        /** Constant for newline separator specification:
            CarriageReturn-Linefeed. */
        public static final int NEWLINE_CR_LF = 1;
        /** Constant for newline separator specification: Use system default. */
        public static final int NEWLINE_DEFAULT = -1;

        // settings
        private boolean skipAnnotations = false;
        private int indent = 3;
        private String newLine;
        private String encoding = null;
        private int lineLength = 0;
        private boolean validate = false;
        private boolean includeBlankLines = true;
        private boolean debug = false;

        // these are used if user specified file input/output
        String inFile = null;
        String outFile = null;

        private XSCtoXML()
        {
		newLine = System.getProperty("line.separator");
		XSCWarning.resetWarnings();
        }


        private XSCtoXML(boolean skipAnnotations,
                         boolean validateInputFile,
                         int indentation,
                         int newline,
                         int lineLength,
                         boolean includeBlank,
                         String encoding)
        {
                this.skipAnnotations = skipAnnotations;
                this.validate = validateInputFile;

                this.indent = indentation;

		this.newLine = System.getProperty("line.separator");
                if (newline == NEWLINE_LF) this.newLine = "\n";
                else if (newline == NEWLINE_CR) this.newLine = "\r";
                else if (newline == NEWLINE_CR_LF) this.newLine = "\r\n";

                this.encoding = encoding;
                this.lineLength = lineLength;

                this.includeBlankLines = includeBlank;
        }


        // called only from main(); OK to System.exit()
        private void processArguments(String[] args)
        {
		for (int i = 0; i < args.length; i++)
		{
			if (args[i].equals("-d"))
			{ debug = true; continue; }
			if (args[i].equals("-a"))
			{ skipAnnotations = true; continue; }
			if (args[i].equals("-i"))
			{
				if (i + 1 >= args.length)
				{ usage(); System.exit(1); }
				
				i++;
                                int n = 0;
				try { n = Integer.parseInt(args[i]); }
				catch (NumberFormatException e) { usage(); System.exit(1); }
				if (n < 1) { usage(); System.exit(1); }
                                indent = n; continue; }
			if (args[i].equals("-v"))
			{ validate = true; continue; }
			if (args[i].equals("-e"))
			{
				if (i + 1 >= args.length) { usage(); System.exit(1); }
				encoding = args[++i];
				continue;
			}
			if (args[i].equals("-c"))
			{ includeBlankLines = false; continue; }

			if (args[i].equals("-n"))
			{
				if (i + 1 >= args.length)
				{ usage(); System.exit(1); }
				
				i++;
				if (args[i].equals("LF"))  newLine = "\n";
				else if (args[i].equals("CR"))  newLine = "\r";
				else if (args[i].equals("CR-LF")) newLine = "\r\n";
				else { usage(); System.exit(1); }
				continue;
			}
			
			if (args[i].equals("-w"))
			{
				if (i + 1 >= args.length)
				{ usage(); System.exit(1); }
				
				int n = 0;
				try { n = Integer.parseInt(args[++i]); }
				catch (NumberFormatException e) { usage(); System.exit(1); }
				
				if (n < 20) { usage(); System.exit(1); }
				lineLength = n;
				continue;
			}

			if (inFile == null) inFile = args[i];
			else if (outFile == null) outFile = args[i];
			else { usage(); System.exit(1); }
		}
		if (inFile == null) { usage(); System.exit(1); }
		if (outFile == null) outFile = addExtension(inFile);
        }


        private Document parseInputFile(String inFile)
            throws IOException
        {
		InputStreamReader reader = null;
		try {
			FileInputStream is = new FileInputStream(inFile);
			if (encoding == null)
				reader = new InputStreamReader(is);
			else
				reader = new InputStreamReader(is,encoding);
		} catch (Exception e)
		{
                        if (debug) e.printStackTrace();
			System.err.println(e.getMessage());
			System.exit(1);
		}
                return parseInput(reader);
        }

        private Document parseInputString(String input)
        {
                StringReader reader = new StringReader(input);
                return parseInput(reader);
        }

        private Document parseInput(Reader reader)
        {
		XSCDOMBuilder builder= new XSCDOMBuilder(debug,skipAnnotations);
		Document doc = null;
		try {
			doc = builder.parse(reader,new DOMImplementationImpl());
		} catch (Exception e)
		{
                        if (debug) {
                                e.printStackTrace();
                                System.err.println(e.getMessage());
                        }
			System.exit(1);
		}
                return doc;
        }


        private void writeOutputFile(Document inDoc, String outFile)
            throws IOException
        {
		FileWriter fileWriter = null;
		try {
			fileWriter = new FileWriter(outFile);
		} catch (Exception e)
		{
			System.err.println(e.getMessage());
			System.exit(1);
		}

                String output = writeOutput(inDoc);

                try {
                        fileWriter.write(output);
                        fileWriter.close();
		} catch (Exception e)
		{
			System.err.println(e.getMessage());
			System.exit(1);
		}
        }

        private String writeOutput(Document inDoc) throws IOException
        {
                StringWriter stringWriter = new StringWriter();
		OutputFormat format = new OutputFormat(inDoc,encoding,
                                                       indent != 0);
                if (indent != 0)
                    format.setIndent(indent);
                format.setLineWidth(lineLength);
                format.setLineSeparator(newLine);

		XMLSerializer ser = new XMLSerializer(stringWriter,format);
		XSCWarning.resetWarnings();
                ser.serialize(inDoc);
                
                String result = stringWriter.toString();
                if (includeBlankLines) {
                        String indentStr = "";
                        for (int i=0; i<indent; i++)
                                indentStr += " ";
                        result = result.replaceAll("(?m)^("+indentStr+"<[^>/]+>)$",
                                                   newLine+"$1");
                }
                return result;
        }


        /**
         *  API method: String input, String output; takes arguments directly.
         *  See {@link #convert(boolean,boolean,int,int,String,int)} for
         *  information on default settings.
         */
        public static String convert(String input) throws IOException

        {
            return convert(input, false,false,3, NEWLINE_DEFAULT, 0,true,null);
        }

        /**
         *  API method: DOM input, String output; takes arguments directly.
         *  
         *  @param skipAnnotations  Don't convert annotations (false)
         *  @param validateInputFile report whether schema well-formed (false)
         *  @param indentation  indent unit (#/spaces); 0=none (3)
         *  @param newline  {NEWLINE_LF,NEWLINE_CR,NEWLINE_CR_LF} for Unix,
         *                  Mac, or Windows; special value NEWLINE_DEFAULT will
         *                  select platform default (NEWLINE_DEFAULT)
         *  @param encoding specify encoding to use, null for platform default
         *                  (null)
         *  @param lineLength wrap lines after this many characters,
         *                    0 means don't wrap (0)
         *  @param includeBlank include blank lines in output for readability
         *                      (true)
         */
        public static String convert(String input,
                                     boolean skipAnnotations,
                                     boolean validateInputFile,
                                     int indentation,
                                     int newline,
                                     int lineLength,
                                     boolean includeBlank,
                                     String encoding)
            throws IOException
        {
                // init
                XSCtoXML converter = new XSCtoXML
                    (skipAnnotations, validateInputFile, indentation, newline,
                     lineLength, includeBlank, encoding);

                // go
                Document doc = converter.parseInputString(input);
                String result = converter.writeOutput(doc);

                return result;
        }


        /**
         *  Command-line method: file input/output; reads arguments as switches.
         */
	public static void main(String[] args) throws IOException
	{
                XSCtoXML converter = new XSCtoXML();
                converter.processArguments(args);

                Document doc = converter.parseInputFile(converter.inFile);
                converter.writeOutputFile(doc, converter.outFile);
		XSCWarning.printWarnings(System.out);

		if (converter.validate) validate(converter.outFile);
	}


	private static void validate(String file)
	{
		XMLSchemaLoader loader = new XMLSchemaLoader();
		FileReader reader = null;
		try {
			reader = new FileReader(file);
		} catch (Exception e)
		{
			System.err.println("Validation: " + e.getMessage());
			System.exit(1);
		}
		XMLInputSource source = new XMLInputSource(null,file,null,reader,null);

		try {
			Grammar grammar = loader.loadGrammar(source);
		} catch (Exception e)
		{
			System.err.println("Validation: " + e.getMessage());
			System.exit(1);
		}
	}


	private static String addExtension(String inFile)
	{
		// find last dot
		String namepart = null;

		int dotIndex = inFile.lastIndexOf('.');
		if (dotIndex <= 0) namepart = inFile;
		else namepart = inFile.substring(0,dotIndex);

		// check if exists
		File f = new File(namepart + ".xsd");
		if (!f.exists()) return (namepart + ".xsd");

		// create another filename
		String newNamepart = null;
		for (int i = 1;true;i++)
		{
			f = new File(namepart + i + ".xsd");
			if (!f.exists()) return (namepart + i + ".xsd");
		}
	}


	private static void usage()
	{
		System.out.println("usage: java noown.domain.xsc.util.XSCtoXML [options] infile [outfile]");
		System.out.println("options: -a : ignore annotations (false)");
		System.out.println("         -v : validate file using xerces (false)");
		System.out.println("         -i : output indent unit (0 for none; default 3)");
		System.out.println("         -n newline: specify newline (CR, CR-LF or LF)");
		System.out.println("         -w number: wrap lines after number characters");
		System.out.println("         -c: compact format: do not insert blank lines between definitions");
		System.out.println("         -e encoding : specify encoding of input/output file");
		System.out.println("         -d : debug output");
	}
}
