;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;   A GNU Emacs / XEmacs major mode for editing XML Schema Compact Syntax,
;;   a parallel to RELAX NG Compact syntax available for XML Schema
;;   (see http://dret.net/projects/xscs/)
;;
;;   Place in your site-lisp directory and put the following in ~/.emacs:
;;
;;   (autoload 'xsc-mode "xsc-mode")
;;   (setq auto-mode-alist (cons '("\\.xsc\\'" . xsc-mode) auto-mode-alist))
;;
;;   v 0.5, 2004/08/10
;;   Adrian Robert, based on rnc-mode 1.0b3 of 2002-12-05.
;;   BSD-style license notice of rnc-mode follows; no additional
;;   restrictions imposed for xsc-mode beyond those it declares.
;;   See http://www.pantor.com/download.html .
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;   Copyright (c) 2002, Pantor Engineering AB
;;   All rights reserved.
;;   
;;   Redistribution and use in source and binary forms, with or
;;   without modification, are permitted provided that the following
;;   conditions are met:
;;   
;;   * Redistributions of source code must retain the above copyright
;;     notice, this list of conditions and the following disclaimer. 
;;
;;   * Redistributions in binary form must reproduce the above
;;     copyright notice, this list of conditions and the following
;;     disclaimer in the documentation and/or other materials provided
;;     with the distribution.
;;
;;   * Neither the name of Pantor Engineering AB nor the names of its
;;     contributors may be used to endorse or promote products derived
;;     from this software without specific prior written permission.
;;   
;;   THIS   SOFTWARE   IS PROVIDED BY    THE   COPYRIGHT  HOLDERS  AND
;;   CONTRIBUTORS "AS IS"   AND ANY  EXPRESS OR  IMPLIED   WARRANTIES,
;;   INCLUDING,  BUT  NOT LIMITED  TO,   THE  IMPLIED  WARRANTIES   OF
;;   MERCHANTABILITY    AND  FITNESS  FOR   A  PARTICULAR  PURPOSE ARE
;;   DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS
;;   BE  LIABLE   FOR ANY    DIRECT, INDIRECT,   INCIDENTAL,  SPECIAL,
;;   EXEMPLARY, OR CONSEQUENTIAL DAMAGES  (INCLUDING, BUT NOT  LIMITED
;;   TO, PROCUREMENT  OF  SUBSTITUTE GOODS OR  SERVICES;  LOSS OF USE,
;;   DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
;;   ANY THEORY OF  LIABILITY, WHETHER IN CONTRACT,  STRICT LIABILITY,
;;   OR  TORT (INCLUDING NEGLIGENCE OR  OTHERWISE) ARISING  IN ANY WAY
;;   OUT OF  THE  USE OF   THIS  SOFTWARE,  EVEN IF ADVISED   OF   THE
;;   POSSIBILITY OF SUCH DAMAGE.
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;   Created by David.Rosenborg@pantor.com
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;   Example setup for your ~/.emacs file:
;;
;;   (autoload 'rnc-mode "rnc-mode")
;;   (setq auto-mode-alist       
;;         (cons '("\\.rnc\\'" . rnc-mode) auto-mode-alist))
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;   Changes since 1.0b:
;;     Added a couple of defvars for faces to handle differences
;;     between GNU Emacs and XEmacs.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(require 'font-lock)

(defvar xsc-indent-level 3 "The XSC indentation level.")
  
(defvar xsc-keywords
  (mapcar (lambda (kw) (concat "\\b" kw "\\b"))
	  '("targetNamespace" "namespace" "import"
	    "simpleType" "complexType" "attribute" "element" "required"
	    "list"))
  "XSC keywords")

(defvar xsc-atoms
  (mapcar (lambda (kw) (concat "\\b" kw "\\b"))
	  '("empty" "notAllowed" "string" "text" "token"))
  "XSC atomic pattern keywords")

(defun xsc-make-regexp-choice (operands)
  "(op1 op2 ...) -> \"\\(op1\\|op2\\|...\\)\""
  (let ((result "\\("))
    (mapcar (lambda (op) (setq result (concat result op "\\|"))) operands)
    (concat (substring result 0 -2) "\\)")))

;; Font lock treats face names differently in GNU Emacs and XEmacs
;; The following defvars is a workaround

(defvar italic 'italic)
(defvar default 'default)
(defvar font-lock-comment-face 'font-lock-comment-face)
(defvar font-lock-function-name-face 'font-lock-function-name-face)
(defvar font-lock-preprocessor-face 'font-lock-preprocessor-face)
(defvar font-lock-string-face 'font-lock-string-face)

(defvar xsc-font-lock-keywords
  (list
   '("/[^*/].*/" . font-lock-string-face)
   '("\\b\\(attribute\\|element\\|simpleType\\|complexType\\)\\b\\([- :a-zA-Z0-9._]+\\)" 2 font-lock-variable-name-face)
   '("[a-zA-Z][-a-zA-Z0-9._]*:[a-zA-Z][-a-zA-Z0-9._]*" . italic)
; first one of these puts namespace name if found in blue; second makes
; 'namespace' or 'targetNamespace' pink
   '("\\(namespace\\|targetNamespace\\)\\(\\s +[a-zA-Z][-a-zA-Z0-9._]*\\)\\s *" 2 font-lock-function-name-face)
   '("\\(namespace\\|targetNamespace\\)\\(\\s +[a-zA-Z][-a-zA-Z0-9._]*\\)?\\s *" 1 font-lock-preprocessor-face)

   '("[a-zA-Z][a-zA-Z0-9._]*\\(-[a-zA-Z][a-zA-Z0-9._]*\\)+" . default)
   (cons (xsc-make-regexp-choice xsc-atoms) 'italic)
   (cons (xsc-make-regexp-choice xsc-keywords) font-lock-keyword-face)
   )
  "XSC Highlighting")

;   '("\\([a-zA-Z][-a-zA-Z0-9._]*\\)\\(\\s \\|\n\\)*[|&]?=" 1 
;     font-lock-function-name-face)


(defun xsc-find-column (first start)
  "Find which column to indent to." 

  ;; FIXME: backward-sexp doesn't work with unbalanced braces in comments

  (let* (column
	 pos
	 ;; Find start of enclosing block or assignment
	 (token
	  (if (member first '("]" "}" ")"))
	      (progn
		(goto-char (+ start 1))
		(backward-sexp)
		(beginning-of-line)
		(re-search-forward "\\S ")
		(setq pos (point))
		(setq column (- (current-column) 1))
		'lpar)
	    (catch 'done
	      (while (setq pos (re-search-backward "[{}()=]\\|\\[\\|\\]" 
						   (point-min) t))
		(let ((c (match-string 0)))
		  (beginning-of-line)
		  (re-search-forward "\\S ")
		  (setq column (- (current-column) 1))
		  (beginning-of-line)
		  (cond
		   ;; Don't match inside comments
		   ;; FIXME: Should exclude matches inside string literals too
		   ((re-search-forward "#" pos t) (beginning-of-line))
		   ;; Skip block
		   ((member c '("]" "}" ")"))
		    (goto-char (+ pos 1))
		    (backward-sexp))

		   ((string= c "=") (throw 'done 'eq))
		   (t (throw 'done 'lpar)))))))))

    (cond
     ((not pos) 0)
     ((member first '("]" "}" ")")) column)
     ((member first '("{" "(")) (+ column xsc-indent-level))

     ;; Give lines starting with an operator a small negative indent.
     ;; This allows for the following indentation style:
     ;;   foo =
     ;;      bar
     ;;    | baz
     ;;    | oof
     ((member first '("," "&" "|")) (+ column (- xsc-indent-level 2)))

     ;; Check if first preceding non-whitespace character was an operator
     ;; If not, this is most likely a new assignment.
     ;; FIXME: This doesn't play well with name classes starting on a new
     ;; line
     ((eq token 'eq)
      (goto-char start)
      (if (and (re-search-backward "[^ \t\n]" (point-min) t)
	       (member (match-string 0) '("&" "|" "," "=" "~")))
	  (+ column xsc-indent-level)
	column))

     (t (+ column xsc-indent-level)))))

(defun xsc-indent-line ()
  "Indents the current line."
  (interactive)
  (let ((orig-point (point)))
    (beginning-of-line)
    (let* ((beg-of-line (point))
	   (pos (re-search-forward "\\(\\S \\|\n\\)" (point-max) t))
	   (first (match-string 0))
	   (start (match-beginning 0))
	   (col (- (current-column) 1)))

      (goto-char beg-of-line)

      (let ((indent-column (xsc-find-column first start)))
	(goto-char beg-of-line)

	(cond
	 ;; Only modify buffer if the line must be reindented
	 ((not (= col indent-column))
	  (if (not (or (null pos)
		       (= beg-of-line start)))
	      (kill-region beg-of-line start))

	  (goto-char beg-of-line)
	
	  (while (< 0 indent-column)
	    (insert " ")
	    (setq indent-column (- indent-column 1))))

	 ((< orig-point start) (goto-char start))
	 (t (goto-char orig-point)))))))
	

(defun xsc-electric-brace (arg)
  (interactive "*P")
  (self-insert-command (prefix-numeric-value arg))
  (xsc-indent-line)
  (let ((p (point)))
    (when (save-excursion
	    (beginning-of-line)
	    (let ((pos (re-search-forward "\\S " (point-max) t)))
	      (and pos (= (- pos 1) p))))
      (forward-char))))

(defvar xsc-mode-map () "Keymap used in XSC mode.")
(when (not xsc-mode-map)
  (setq xsc-mode-map (make-sparse-keymap))
  (define-key xsc-mode-map "\C-c\C-c" 'comment-region)
  (define-key xsc-mode-map "}" 'xsc-electric-brace)
  (define-key xsc-mode-map "{" 'xsc-electric-brace)
  (define-key xsc-mode-map "]" 'xsc-electric-brace)
  (define-key xsc-mode-map "[" 'xsc-electric-brace))

(defun xsc-mode ()
  "Major mode for editing RELAX NG Compact Syntax schemas.
\\{xsc-mode-map}"
  (interactive)

  (kill-all-local-variables)

  (make-local-variable 'indent-line-function)
  (setq indent-line-function 'xsc-indent-line)
  
  (make-local-variable 'font-lock-defaults)
  (setq font-lock-defaults '(xsc-font-lock-keywords nil t nil nil))
  
  (use-local-map xsc-mode-map)

  (make-local-variable 'comment-start)
  (make-local-variable 'comment-end)
  (make-local-variable 'comment-start-skip)

;;   (setq comment-start "#"
;; 	comment-end ""
;; 	comment-start-skip "\\([ \n\t]+\\)##?[ \n\t]+")
  (setq comment-start "/*"
	comment-end "*/"
	comment-start-skip "\\([ \n\t]+\\)##?[ \n\t]+")

  (let ((xsc-syntax-table (copy-syntax-table)))
    (modify-syntax-entry ?/ ". 14 " xsc-syntax-table)
    (modify-syntax-entry ?* ". 23 " xsc-syntax-table)
    (modify-syntax-entry ?\\ "w   " xsc-syntax-table)
    (modify-syntax-entry ?' "\"   " xsc-syntax-table)
; Below was causing conflicts w/comments; do it in keyword-highlighting instead.
;;    (modify-syntax-entry ?/ "\"   " xsc-syntax-table)
    (modify-syntax-entry ?. "w   " xsc-syntax-table)
    (modify-syntax-entry ?- "w   " xsc-syntax-table)
    (modify-syntax-entry ?_ "w   " xsc-syntax-table)
    (set-syntax-table xsc-syntax-table))
  
  (setq mode-name "XSC"
	major-mode 'xsc-mode)
  (run-hooks 'xsc-mode-hook))

(provide 'xsc-mode)
